from pydantic import BaseModel
from pydantic_resolve.utils.er_diagram import base_entity, Relationship

Base = base_entity()

class Sub(BaseModel):
    id: int
    name: str

class MyModel(BaseModel, Base):
    __pydantic_resolve_relationships__ = [
        Relationship(field='id', target_kls=Sub, loader=None)
    ]
    id: int

class MyModelSub(MyModel):
    id: int

class AnotherModel(BaseModel, Base):
    __pydantic_resolve_relationships__ = [
        Relationship(field='id', target_kls=Sub, loader=None)
    ]
    id: int

def test_bases():
    assert MyModel in Base.entities
    assert AnotherModel in Base.entities
    assert MyModelSub not in Base.entities
    assert len(Base.entities) == 2



from dataclasses import dataclass
from typing import Iterator, Tuple, Type
import pydantic_resolve.constant as const

@dataclass
class ExposeInfo:
    alias: str

def ExposeAs(alias: str) -> ExposeInfo:
    return ExposeInfo(alias=alias)


def pre_generate_expose_config(kls):
    """
    iterrate kls fields, check and collect field who's annotated metadata for ExposeAs exists
    if kls's const.EXPOSE_CONFIGURATION exists and the fields is not empty, raise exception

    generate the configuration such as 
    { field_name: alias }
    and set it into kls's const.EXPOSE_CONFIGURATION
    """
    fields = list(_get_pydantic_field_items_with_expose_as(kls))
    if not fields:
        return

    if hasattr(kls, const.EXPOSE_TO_DESCENDANT):
        raise AttributeError(
            f"{const.EXPOSE_TO_DESCENDANT} already exists; cannot use ExposeAs annotations at the same time"
        )

    expose_dict = {name: meta.alias for name, meta in fields}
    setattr(kls, const.EXPOSE_TO_DESCENDANT, expose_dict)

def _get_pydantic_field_items_with_expose_as(kls) -> Iterator[Tuple[str, ExposeInfo, Type]]:
    items = kls.model_fields.items()

    for name, v in items:
        metadata = v.metadata
        for meta in metadata:
            if isinstance(meta, ExposeInfo):
                yield name, meta


from uuid import uuid1
from pydantic_resolve import Resolver
import gc
from pydantic import BaseModel
import psutil
import pytest


class Model(BaseModel):
    id: int

    compound_id: str = ''
    def post_compound_id(self, parent, ancestor_context):
        return f'{parent.id}-{self.id}-{ancestor_context.get("parent_id")}'

class Parent(BaseModel):
    __pydantic_resolve_expose__ = {'id': 'parent_id'}
    id: int
    m: Model
    


def str_uuid1():
    return str(uuid1())

@pytest.mark.asyncio
async def test_mem_leak_2():
    process = psutil.Process()
    gc.collect()

    memory_start = process.memory_info().rss
    for _ in range(10000):
        await Resolver().resolve(Parent(id=1, m=Model(id=2)))

    memory_after_del = process.memory_info().rss
    diff = memory_after_del - memory_start
    print(memory_after_del, memory_start, diff/1024)
    assert diff/1024 < 500  # less than 500KB increase


from pydantic import BaseModel
from typing import List
from pydantic_resolve import Resolver, LoaderDepend
from aiodataloader import DataLoader
import pytest

class LoaderA(DataLoader):
    power: int
    async def batch_load_fn(self, keys: List[int]):
        return [ k** self.power for k in keys ]

class LoaderB(DataLoader):
    power: int
    async def batch_load_fn(self, keys: List[int]):
        return [ k** self.power for k in keys ]

class LoaderC(DataLoader):
    power: int
    add: int 
    async def batch_load_fn(self, keys: List[int]):
        return [ k** self.power + self.add for k in keys ]


async def loader_fn_a(keys):
    return [ k**2 for k in keys ]

class A(BaseModel):
    val: int

    a: int = 0
    def resolve_a(self, loader=LoaderDepend(LoaderA)):
        return loader.load(self.val)

    b: int = 0
    def resolve_b(self, loader=LoaderDepend(LoaderB)):
        return loader.load(self.val)

    c: int = 0
    def resolve_c(self, loader=LoaderDepend(LoaderC)):
        return loader.load(self.val)


@pytest.mark.asyncio
async def test_case_0():
    data = [A(val=n) for n in range(3)]
    with pytest.warns(DeprecationWarning):
        data = await Resolver(global_loader_filter={'power': 2}, 
                            loader_filters={LoaderC:{'add': 1}}).resolve(data)


@pytest.mark.asyncio
async def test_case_1():
    data = [A(val=n) for n in range(3)]
    with pytest.warns(DeprecationWarning):
        data = await Resolver(loader_filters={LoaderA:{'power': 2}, 
                                            LoaderB:{'power': 3},
                                            LoaderC:{'power': 3, 'add': 1}}).resolve(data)




from dataclasses import dataclass
from typing import Optional
from pydantic import BaseModel
from pydantic_resolve.utils import class_util
import pydantic_resolve.constant as const
import pytest

def test_ensure_subset():
    class Base(BaseModel):
        a: str
        b: int

    @class_util.ensure_subset(Base)
    class ChildA(BaseModel):
        a: str

    @class_util.ensure_subset(Base)
    class ChildB(BaseModel):
        a: str
        c: int = 0
        def resolve_c(self):
            return 21

    @class_util.ensure_subset(Base)
    class ChildB1(BaseModel):
        a: str
        d: Optional[int] = 0

    # Test that ENSURE_SUBSET_REFERENCE is set correctly
    assert hasattr(ChildA, const.ENSURE_SUBSET_REFERENCE)
    assert getattr(ChildA, const.ENSURE_SUBSET_REFERENCE) is Base
    assert hasattr(ChildB, const.ENSURE_SUBSET_REFERENCE)
    assert getattr(ChildB, const.ENSURE_SUBSET_REFERENCE) is Base
    assert hasattr(ChildB1, const.ENSURE_SUBSET_REFERENCE)
    assert getattr(ChildB1, const.ENSURE_SUBSET_REFERENCE) is Base

    with pytest.raises(AttributeError):
        @class_util.ensure_subset(Base)
        class ChildC(BaseModel):
            a: str
            b: str

    with pytest.raises(AttributeError):
        @class_util.ensure_subset(Base)
        class ChildD(BaseModel):
            a: str
            b: int
            c: int

    with pytest.raises(TypeError):
        @class_util.ensure_subset(Base)
        @dataclass
        class ChildE():
            a: str
            b: int
            c: int




from typing import Any, Dict
from pydantic import BaseModel
from pydantic_resolve.utils.class_util import safe_issubclass


def model_config(default_required: bool=True):
    """
    in pydantic v2, we can not use __exclude_field__ to set hidden field in model_config hidden_field params
    model_config now is just a simple decorator to remove fields (with exclude=True) from schema.properties
    and set schema.required for better schema description. 
    (same like `output` decorator, you can replace output with model_config)

    it keeps the form of model_config(params) in order to extend new features in future

    update:

    in fastapi + pydantic v2, this function will be handled internal automatically with mode: serilization
    you can remove the model_config_v2

    reference: fastapi/_compat.py::get_definitions
    """
    def wrapper(kls):
        if safe_issubclass(kls, BaseModel):
            # TODO: check the behavior of generating json schema in other frameworks using pydantic
            def build():
                def _schema_extra(schema: Dict[str, Any], model) -> None:
                    # 1. collect exclude fields and then hide in both schema and dump (default action)
                    excluded_fields = [k for k, v in kls.model_fields.items() if v.exclude]
                    props = {}

                    # config schema properties
                    for k, v in schema.get('properties', {}).items():
                        if k not in excluded_fields:
                            props[k] = v
                    schema['properties'] = props

                    # config schema required (fields with default values will not be listed in required field)
                    # and the generated typescript models will define it as optional, and is troublesome in use
                    if default_required:
                        fnames = kls.model_fields.keys()
                        if excluded_fields:
                            fnames = [n for n in fnames if n not in excluded_fields]
                        schema['required'] = fnames

                return _schema_extra

            kls.model_config['json_schema_extra'] = staticmethod(build())
            return kls
        else:
            raise AttributeError(f'target class {kls.__name__} is not BaseModel')
    return wrapper




import pytest
from pydantic import BaseModel, ValidationError

from pydantic_resolve.utils.er_diagram import (
	Entity,
	ErDiagram,
	Link,
	MultipleRelationship,
	Relationship,
)


class A(BaseModel):
	fk: int


class T1(BaseModel):
	id: int


class T2(BaseModel):
	id: int


def dummy_loader(*args, **kwargs):
	return None


def test_relationship_single_target_no_biz_required():
	cfg = Entity(
		kls=A,
		relationships=[
			Relationship(field="fk", target_kls=T1, loader=dummy_loader),
		],
	)
	assert isinstance(cfg, Entity)


def test_relationship_none_default_conflict_raises():
	with pytest.raises(ValidationError):
		Relationship(
			field="fk",
			target_kls=T1,
			loader=dummy_loader,
			field_none_default=None,
			field_none_default_factory=list,
		)


def test_entity_duplicate_field_target_raises():
	with pytest.raises(ValidationError):
		Entity(
			kls=A,
			relationships=[
				Relationship(field="fk", target_kls=T1, loader=dummy_loader),
				Relationship(field="fk", target_kls=T1, loader=dummy_loader),
			],
		)


def test_entity_duplicate_multiple_relationship_raises():
	with pytest.raises(ValidationError):
		Entity(
			kls=A,
			relationships=[
				MultipleRelationship(
					field="fk",
					target_kls=list[T1],
					links=[Link(biz="a", loader=dummy_loader)],
				),
				MultipleRelationship(
					field="fk",
					target_kls=list[T1],
					links=[Link(biz="b", loader=dummy_loader)],
				),
			],
		)

def test_multiple_relationship_unique_links_allowed():
	entity = Entity(
		kls=A,
		relationships=[
			MultipleRelationship(
				field="fk",
				target_kls=list[T1],
				links=[
					Link(biz="a", loader=dummy_loader),
					Link(biz="b", loader=dummy_loader),
				],
			)
		],
	)
	assert isinstance(entity, Entity)

def test_multiple_relationship_duplicated_disallowed():
	with pytest.raises(ValidationError):
		Entity(
			kls=A,
			relationships=[
				MultipleRelationship(
					field="fk",
					target_kls=list[T1],
					links=[
						Link(biz="a", loader=dummy_loader),
						Link(biz="a", loader=dummy_loader),
					],
				)
			],
		)

def test_entity_allows_distinct_targets_per_field():
	entity = Entity(
		kls=A,
		relationships=[
			Relationship(field="fk", target_kls=T1, loader=dummy_loader),
			Relationship(field="fk", target_kls=T2, loader=dummy_loader),
		],
	)
	assert isinstance(entity, Entity)


def test_erdiagram_duplicate_entity_kls_raises():
	cfg = Entity(
		kls=A,
		relationships=[Relationship(field="fk", target_kls=T1, loader=dummy_loader)],
	)
	with pytest.raises(ValidationError):
		ErDiagram(configs=[cfg, cfg])


from inspect import isclass
from typing import List, Type

import pydantic_resolve.utils.class_util as class_util
import pydantic_resolve.utils.params as params_util
from pydantic_resolve.analysis import LoaderQueryMeta, MappedMetaType
from pydantic_resolve.exceptions import LoaderFieldNotProvidedError

from aiodataloader import DataLoader
from pydantic import BaseModel


class LoaderManager:
    """
    Refactor loader instance validation/creation into a class.

    Public API:
      - validate_and_create_loader_instance(self, loader_params, global_loader_param, loader_instances, metadata)
    """

    def __init__(self) -> None:
        pass

    # fetch all loaders
    def _get_all_loaders_from_meta(self, metadata: MappedMetaType):
        for _, kls_info in metadata.items():
            for _, resolve_info in kls_info['resolve_params'].items():
                for loader in resolve_info['dataloaders']:
                    yield loader

            for _, post_info in kls_info['post_params'].items():
                for loader in post_info['dataloaders']:
                    yield loader

    def _create_instance(self, loader, loader_params, global_loader_param, loader_instances):
        """
        1. is class?
            - validate params
        2. is func
        """
        loader_kls, path = loader['kls'], loader['path']
        if isclass(loader_kls):
            loader_instance = loader_kls()
            param_config = params_util.merge_dicts(
                global_loader_param,
                loader_params.get(loader_kls, {}))

            for field, has_default in class_util.get_fields_default_value_not_provided(loader_kls):
                try:
                    if has_default and field not in param_config:
                        continue

                    value = param_config[field]
                    setattr(loader_instance, field, value)
                except KeyError:
                    raise LoaderFieldNotProvidedError(f'{path}.{field} not found in Resolver()')
            return loader_instance
        else:
            return DataLoader(batch_load_fn=loader_kls)  # type:ignore

    def _get_all_fields(self, kls: Type):
        if class_util.safe_issubclass(kls, BaseModel):
            return list(class_util.get_pydantic_field_keys(kls))
        else:
            raise AttributeError('invalid type: should be pydantic object')  # noqa

    def _generate_meta(self, types: List[List[Type]]):
        _fields = set()
        meta: LoaderQueryMeta = {
            'fields': [],
            'request_types': []
        }

        for tt in types:
            for t in tt:
                fields = self._get_all_fields(t)
                meta['request_types'].append(dict(name=t, fields=fields))
                _fields.update(fields)
        meta['fields'] = list(_fields)
        return meta

    def validate_and_create_loader_instance(
        self,
        loader_params,
        global_loader_param,
        loader_instances,
        metadata: MappedMetaType
    ):
        """
        return loader_instance_cache

        validate: whether loader params are missing
        create:
            - func
            - loader class
                - no param
                - has param
        """
        cache = {}
        request_info = {}

        # create instance
        for loader in self._get_all_loaders_from_meta(metadata):
            loader_kls, path = loader['kls'], loader['path']

            if path in cache:
                continue

            if loader_instances.get(loader_kls):  # if instance already exists
                cache[path] = loader_instances.get(loader_kls)
                continue

            cache[path] = self._create_instance(loader, loader_params, global_loader_param, loader_instances)

        # prepare query meta
        for loader in self._get_all_loaders_from_meta(metadata):
            kls, path = loader['request_type'], loader['path']

            if kls is None:
                continue

            if path in request_info and kls not in request_info[path]:
                request_info[path].append(kls)
            else:
                request_info[path] = [kls]

        # combine together
        for path, instance in cache.items():
            if request_info.get(path) is None:
                continue
            instance._query_meta = self._generate_meta(request_info[path])

        return cache


def validate_and_create_loader_instance(
        loader_params,
        global_loader_param,
        loader_instances,
        metadata: MappedMetaType):

    return LoaderManager().validate_and_create_loader_instance(
        loader_params=loader_params,
        global_loader_param=global_loader_param,
        loader_instances=loader_instances,
        metadata=metadata,
    )


import asyncio
from pydantic import ConfigDict, BaseModel
import pytest

import pydantic_resolve.utils.class_util
import pydantic_resolve.utils.conversion
import pydantic_resolve.utils.dataloader

def test_get_class_field_annotations():
    class B:
        hello: str = 'hello'

    class C:
        hello: str

        def __init__(self, c: str):
            self.c = c
        
    class D(C):
        pass

    class E(C):
        world: str
    
    assert list(pydantic_resolve.utils.class_util.get_fields_default_value_not_provided(B)) == [('hello', True)]
    assert list(pydantic_resolve.utils.class_util.get_fields_default_value_not_provided(C)) == [('hello', False)]
    assert list(pydantic_resolve.utils.class_util.get_fields_default_value_not_provided(D)) == []
    assert list(pydantic_resolve.utils.class_util.get_fields_default_value_not_provided(E)) == [('world', False)]


class User(BaseModel):
    id: int
    name: str
    age: int


def test_build_object():
    raw = [(1, 'peter', 10), (2, 'mike', 21), (3, 'john', 12)]
    users = [User(id=i[0], name=i[1], age=i[2]) for i in raw]
    a, b, c = users
    ids = [2, 3, 1, 4]
    output = pydantic_resolve.utils.dataloader.build_object(users, ids, lambda x: x.id)
    assert list(output) == [b, c, a, None]
    

def test_build_list():
    raw = [(1, 'peter', 10), (2, 'mike', 21), (3, 'john', 12)]
    users = [User(id=i[0], name=i[1], age=i[2]) for i in raw]
    a, b, c = users
    ids = [2, 3, 1, 4]
    output = pydantic_resolve.utils.dataloader.build_list(users, ids, lambda x: x.id)
    assert list(output) == [[b], [c], [a], []]



def test_super_logic():
    class A():
        def say(self):
            return 'A'
    
    class B(A):
        def say(self):
            val = A().say()
            return f'B.{val}'
    

    b = B()
    assert b.say() == 'B.A'



@pytest.mark.asyncio
async def test_mapper_1():
    class A(BaseModel):
        a: int

    @pydantic_resolve.utils.conversion.mapper(lambda x: A(**x))
    async def foo():
        return {'a': 1}

    async def call_later(f):
        await asyncio.sleep(.1)
        f.set_result({'a': 1})

    @pydantic_resolve.utils.conversion.mapper(lambda x: A(**x))
    async def bar():
        lp = asyncio.get_event_loop()
        f = lp.create_future()
        asyncio.create_task(call_later(f))
        return f
    
    ret = await foo()
    ret2 = await bar()
    assert ret == A(a=1)
    assert ret2 == A(a=1)


def test_auto_mapper_2():
    class A(BaseModel):
        a: int
        model_config = ConfigDict(from_attributes=True)
    
    class AA:
        def __init__(self, a):
            self.a = a
    
    p1 = (A, AA(1))
    ret = pydantic_resolve.utils.conversion._get_mapping_rule(*p1)(*p1)  # type: ignore
    assert ret == A(a=1)

    p2 = (A, {'a': 1})
    with pytest.raises(AttributeError):
        pydantic_resolve.utils.conversion._get_mapping_rule(*p2)  # type: ignore
    

def test_auto_mapper_3():
    class A(BaseModel):
        a: int
        model_config = ConfigDict(from_attributes=True)
    
    p1 = (A, A(a=1))
    rule = pydantic_resolve.utils.conversion._get_mapping_rule(*p1)  # type: ignore

    assert rule is None
    output = pydantic_resolve.utils.conversion._apply_rule(rule, *p1, is_list=False)
    assert output == A(a=1)


def test_auto_mapper_4():
    class A(BaseModel):
        a: int

    class AA:
        def __init__(self, a):
            self.a = a
    
    p1 = (A, AA(a=1))
    with pytest.raises(AttributeError):
        pydantic_resolve.utils.conversion._get_mapping_rule(*p1)(*p1)  # type: ignore




from __future__ import annotations
from typing import List
import asyncio
from pydantic import BaseModel
from pydantic_resolve import Resolver
import pytest

class Student(BaseModel):
    new_books: List[str] = []    
    async def resolve_new_books(self) -> List[str]:
        await asyncio.sleep(.1)
        return ['book1', 'book2']

    old_books: List[str] = []    
    async def resolve_old_books(self) -> List[str]:
        await asyncio.sleep(.1)
        return ['book1', 'book2']


@pytest.mark.asyncio
async def test_type_definition():
    stu = Student()
    result = await Resolver().resolve(stu)
    expected = {
        'new_books': ['book1', 'book2'],
        'old_books': ['book1', 'book2']
    }
    assert result.model_dump() == expected



from __future__ import annotations
from pydantic import BaseModel
from pydantic_resolve import Resolver, MissingAnnotationError
import pytest

class Student(BaseModel):
    name: str
    intro: str = ''
    def resolve_intro(self):
        return f'hello {self.name}'

@pytest.mark.asyncio
async def test_check_miss_anno():
    stu = Student(name="martin")
    with pytest.raises(MissingAnnotationError):
        await Resolver(ensure_type=True).resolve(stu)



from __future__ import annotations
from pydantic import BaseModel
from typing import List, Annotated
from pydantic_resolve import Resolver, Collector, LoaderDepend, SendTo, DefineSubset, SubsetConfig
import pytest


class SubCollector(Collector):
    def add(self, val):  # replace with your implementation
        print('add')
        self.val.append(val)

async def c_loader_fn(keys):
    return [[C(detail=f'{k}-1'), C(detail=f'{k}-2')] for k in keys]

class A(BaseModel):
    b_list: List[B] = []
    async def resolve_b_list(self):
        return [dict(name='b1'), dict(name='b2')]

    names: List[str] = []
    def post_names(self, collector=SubCollector('b_name')):
        return collector.values()

    items: List[str] = []
    def post_items(self, collector=Collector('b_items', flat=True)):
        return collector.values()

    details: List[str] = []
    def post_details(self, collector=Collector('c_details', flat=True)):
        return collector.values()

    details_nest: List[List[str]] = []
    def post_details_nest(self, 
                          collector=Collector('c_details')):
        return collector.values()

    details_compare: bool = False
    def post_details_compare(self, 
                          collector=Collector('c_details'),
                          collector2=Collector('c_details'),
                          ):
        return collector.values() == collector2.values()

class B(BaseModel):
    name: Annotated[str, SendTo('b_name')] = ''
    items: Annotated[List[str], SendTo('b_items')] = ['x', 'y']

    details: List[str] = []
    def post_details(self, collector=Collector('c_details', flat=True)):
        return collector.values()

    c_list: List[C] = []
    async def resolve_c_list(self, loader=LoaderDepend(c_loader_fn)):
        return loader.load(self.name)

class C(BaseModel):
    detail: str

    details: Annotated[List[str], SendTo('c_details')] = []
    def resolve_details(self):
        return [f'{self.detail}-detail-1', f'{self.detail}-detail-2']


@pytest.mark.asyncio
async def test_collector_1():
    a = A()
    a = await Resolver().resolve(a)
    assert a.names == ['b1', 'b2']
    assert a.items == ['x', 'y', 'x', 'y']
    assert a.details == ['b1-1-detail-1', 'b1-1-detail-2', 'b1-2-detail-1', 'b1-2-detail-2', 'b2-1-detail-1', 'b2-1-detail-2', 'b2-2-detail-1', 'b2-2-detail-2']
    assert a.details_nest == [['b1-1-detail-1', 'b1-1-detail-2'],
                              ['b1-2-detail-1', 'b1-2-detail-2'], 
                              ['b2-1-detail-1', 'b2-1-detail-2'], 
                              ['b2-2-detail-1', 'b2-2-detail-2']]

    assert a.details_compare is True


class X(BaseModel):
    items: list[Y]
    sum: int = 0
    def post_sum(self, collector=Collector('collector')):
        s = 0
        for a, b in collector.values():
            s += (a + b)
        return s


class Y(BaseModel):
    a: Annotated[int, SendTo('collector')]
    b: Annotated[int, SendTo('collector')]

@pytest.mark.asyncio
async def test_collector_2():
    x = X(items=[Y(a=1, b=2), Y(a=3, b=4)])
    x = await Resolver().resolve(x)
    assert x.sum == 10


class U(BaseModel):
    items: list[V]
    sum: int = 0
    def post_sum(self, collector=Collector('collector1')):
        s = 0
        for a, b in collector.values():
            s += (a + b)
        return s

    sum2: int = 0
    def post_sum2(self, collector=Collector('collector2')):
        s = 0
        for a, b in collector.values():
            s += (a + b)
        return s


class V(BaseModel):
    a: Annotated[int, SendTo(('collector1', 'collector2'))]
    b: Annotated[int, SendTo(('collector1', 'collector2'))]

@pytest.mark.asyncio
async def test_collector_3():
    x = U(items=[V(a=1, b=2), V(a=3, b=4)])
    x = await Resolver().resolve(x)
    assert x.sum == 10
    assert x.sum2 == 10


class E(BaseModel):
    items: list[F]
    sum: int = 0
    def post_sum(self, collector=Collector('collector1')):
        s = 0
        for a, b in collector.values():
            s += (a + b)
        return s

class FBase(BaseModel):
    a: int
    b: int

class F(DefineSubset):
    __subset__ = SubsetConfig(
        kls=FBase,
        fields=['a', 'b'],
        send_to=[('a', 'collector1'), ('b', 'collector1')]
    )

@pytest.mark.asyncio
async def test_collector_with_subclass():
    e = E(items=[F(a=1, b=2), F(a=3, b=4)])
    e = await Resolver().resolve(e) 
    assert e.sum == 10



import functools
from typing import Type, List, Tuple, get_origin, get_args
import pydantic_resolve.constant as const
import pydantic_resolve.utils.class_util as class_util
from pydantic_resolve.analysis import is_acceptable_kls
from pydantic_resolve.utils.types import get_type, get_core_types
from pydantic import BaseModel


def ensure_subset(base):
    """
    used with pydantic class to make sure a class's field is 
    subset of target class

    for pydantic v2, subclass with Optional[T] but without default value will raise exception

    @class_util.ensure_subset(Base)
    class ChildB1(BaseModel):
        a: str
        d: Optional[int]
    
    this will raise

    @class_util.ensure_subset(Base)
    class ChildB1(BaseModel):
        a: str
        d: Optional[int] = 0 
    
    this is ok
    """
    def wrap(kls):
        is_base_pydantic = safe_issubclass(base, BaseModel)
        is_kls_pydantic = safe_issubclass(kls, BaseModel)

        setattr(kls, const.ENSURE_SUBSET_REFERENCE, base)

        if is_base_pydantic and is_kls_pydantic:
            @functools.wraps(kls)
            def inner():
                for k, field in kls.model_fields.items():
                    if field.is_required():
                        base_field = base.model_fields.get(k)
                        if not base_field:
                            raise AttributeError(f'{k} not existed in {base.__name__}.')
                        if base_field and base_field.annotation != field.annotation:
                            raise AttributeError(f'type of {k} not consistent with {base.__name__}')
                return kls
            return inner()
        else:
            raise TypeError('base and kls should both be Pydantic BaseModel')
    return wrap


def get_pydantic_field_items(kls):
    return kls.model_fields.items()


def get_pydantic_field_keys(kls) -> str:
    return kls.model_fields.keys()


def get_pydantic_fields(kls):
    items = class_util.get_pydantic_field_items(kls)

    for name, v in items:
        t = get_type(v)
        shelled_types = get_core_types(t)
        allowed_types = [st for st in shelled_types if is_acceptable_kls(st)]
        if allowed_types:
            yield (name, allowed_types)  # type_ is the most inner type


def get_class_of_object(target):
    if isinstance(target, list):
        return target[0].__class__
    else:
        return target.__class__


def update_forward_refs(kls):
    def update_pydantic_forward_refs(pydantic_kls: type[BaseModel]):
        """
        recursively update refs.
        """

        pydantic_kls.model_rebuild()
        setattr(pydantic_kls, const.PYDANTIC_FORWARD_REF_UPDATED, True)

        values = pydantic_kls.model_fields.values()
        for field in values:
            update_forward_refs(field.annotation)
        
    for shelled_type in get_core_types(kls):
        # Only treat as updated if the flag is set on the class itself, not via inheritance
        local_attrs = getattr(shelled_type, '__dict__', {})
        if local_attrs.get(const.PYDANTIC_FORWARD_REF_UPDATED, False):
            continue

        if safe_issubclass(shelled_type, BaseModel):
            update_pydantic_forward_refs(shelled_type)


def get_kls_full_name(kls):
    return f'{kls.__module__}.{kls.__qualname__}'


def get_fields_default_value_not_provided(cls: Type) -> List[Tuple[str, bool]]:  # field name, has default value
    """
    return class field which do not have a default value.

    class MyClass:
        a: int
        b: int = 1

    print(hasattr(MyClass, 'a'))  # False
    print(hasattr(MyClass, 'b'))  # True
    """
    anno = cls.__dict__.get('__annotations__') or {}
    return [(k, hasattr(cls, k)) for k in anno.keys()]


def safe_issubclass(kls, classinfo):
    try:
        return issubclass(kls, classinfo)
    except TypeError:
        return False
    

def is_compatible_type(src_type, target_type) -> bool:
    """
    1. try to shell optional in src_type and then compare with target_type
    2. if src_type is a subset of target_type, or subset of subset (visit recursively until find), return True
    3. if src_type is subclass of target_type, or subclass of subclass (visit recursively until find), return True
    4. Union and UnionType is not compatible
    """
    # Helper: unwrap Optional[X] in source type only
    def unwrap_optional(tp):
        origin = get_origin(tp)
        if origin is None:
            return tp
        # typing.Union and PEP604 (types.UnionType) both have origin of Union/UnionType
        if str(origin) in {"typing.Union", "types.UnionType"}:
            args = [a for a in get_args(tp) if a is not type(None)]
            if len(args) == 1:
                return args[0]
        return tp

    # Helper: detect any non-optional union
    def is_union(tp):
        origin = get_origin(tp)
        return str(origin) in {"typing.Union", "types.UnionType"}

    # Helper: subset ancestry check via ENSURE_SUBSET_REFERENCE chain
    def is_subset_of(src, tgt) -> bool:
        current = src
        while current is not None:
            if current is tgt:
                return True
            current = getattr(current, const.ENSURE_SUBSET_REFERENCE, None)
        return False

    # Helper: list element compatibility
    def is_list_compatible(src, tgt) -> bool:
        src_origin, tgt_origin = get_origin(src), get_origin(tgt)
        if src_origin is list and tgt_origin is list:
            src_args, tgt_args = get_args(src), get_args(tgt)
            if not src_args or not tgt_args:
                return False
            return is_compatible_type(src_args[0], tgt_args[0])
        return False

    # 1) unwrap Optional in src_type
    src = unwrap_optional(src_type)
    tgt = target_type

    # 4) unions (other than Optional case already unwrapped) are incompatible
    if is_union(src) or is_union(tgt):
        return False

    # 3) handle list generics (typing.List and built-in list share origin list)
    if is_list_compatible(src, tgt):
        return True

    # Direct equality
    if src is tgt:
        return True

    # 2) subset chain match
    try:
        if is_subset_of(src, tgt):
            return True
    except Exception:
        pass

    # 3) subclass chain
    try:
        if safe_issubclass(src, tgt):
            return True
    except Exception:
        pass

    return False



import pytest
from pydantic import BaseModel
from typing import Optional
from pydantic_resolve import Resolver
import asyncio

class D(BaseModel):
    name: str = 'kikodo'
    age: int = 0
    def resolve_age(self):
        return 11

class C(BaseModel):
    name: str = ''
    d: D = D()


class B(BaseModel):
    name: str
    c: Optional[C] = None
    async def resolve_c(self) -> Optional[C]:
        await asyncio.sleep(.1)
        return C(name='hello world')

class A(BaseModel):
    b: B

class Z(BaseModel):
    a: A
    resolve_age: int

@pytest.mark.asyncio
async def test_resolve_object():
    s = Z(a=A(b=B(name="kikodo")), resolve_age=21)

    result = await Resolver().resolve(s)
    expected = {
        "a": {
            "b": {
                "name":"kikodo",
                "c": {
                    "name": "hello world",
                    "d": {
                        "name": "kikodo",
                        "age": 11
                    }
                }
            }
        },
        "resolve_age": 21
    }
    assert result.model_dump() == expected



import pytest
from pydantic import BaseModel
from typing import Optional
from pydantic_resolve import Resolver
import asyncio
import time

class DetailA(BaseModel):
    name: str = ''

class ServiceDetail1(BaseModel):
    detail_a: Optional[DetailA] = None

    async def resolve_detail_a(self) -> Optional[DetailA]:
        await asyncio.sleep(.1)
        return DetailA(name='hello world')

    detail_b: str = ''

    async def resolve_detail_b(self) -> str:
        await asyncio.sleep(.1)
        return 'good'

class Service(BaseModel):
    service_detail_1: Optional[ServiceDetail1] = None
    async def resolve_service_detail_1(self) -> Optional[ServiceDetail1]:
        await asyncio.sleep(.1)
        return ServiceDetail1()

    service_detail_1b: Optional[ServiceDetail1] = None
    async def resolve_service_detail_1b(self) -> Optional[ServiceDetail1]:
        await asyncio.sleep(.1)
        return ServiceDetail1()

    service_detail_2: str = ''
    async def resolve_service_detail_2(self) -> str:
        await asyncio.sleep(.1)
        return "detail_2"

    service_detail_3: str = ''
    async def resolve_service_detail_3(self) -> str:
        await asyncio.sleep(.1)
        return "detail_3"

    service_detail_4: str = ''
    async def resolve_service_detail_4(self) -> str:
        await asyncio.sleep(.1)
        return "detail_4"

@pytest.mark.asyncio
async def test_resolve_object():
    t = time.time()
    s = Service()
    result = await Resolver().resolve(s)
    expected = {
        "service_detail_1": {
            "detail_a": {
                "name": "hello world"
            },
            "detail_b": "good"
        },
        "service_detail_1b": {
            "detail_a": {
                "name": "hello world"
            },
            "detail_b": "good"
        },
        "service_detail_2": "detail_2",
        "service_detail_3": "detail_3",
        "service_detail_4": "detail_4",
    }
    assert result.model_dump() == expected
    delta = time.time() - t

    assert delta < 1


import pytest
from typing import List
from pydantic import BaseModel
from pydantic_resolve.resolver import Resolver


@pytest.mark.asyncio
async def test_case_0():
    class Kar(BaseModel):
        name: str

        desc: str = ''
        def resolve_desc(self, ancestor_context):
            return f"{self.name} - {ancestor_context['bar_num']}"


    class Bar(BaseModel):
        __pydantic_resolve_expose__ = {'num': 'bar_num'}

        num: int

        kars: List[Kar] = []
        def resolve_kars(self):
            return [{'name': n} for n in ['a', 'b', 'c']]


    class Foo(BaseModel):
        __pydantic_resolve_expose__ = {'nums': 'bar_num'}

        nums:List[int]
        bars: List[Bar] = []

        def resolve_bars(self):
            return [{'num': n} for n in self.nums]


    foo = Foo(nums=[1,2,3])
    with pytest.raises(ValueError) as e:
        await Resolver().resolve(foo)
    assert 'alias name conflicts, please check' in str(e.value)


@pytest.mark.asyncio
async def test_case_1():
    class Kar(BaseModel):
        name: str

        desc: str = ''
        def resolve_desc(self, ancestor_context):
            return f"{self.name} - {ancestor_context['bar_num']}"


    class Bar(BaseModel):
        __pydantic_resolve_expose__ = ['num', 'bar_num']

        num: int

        kars: List[Kar] = []
        def resolve_kars(self):
            return [{'name': n} for n in ['a', 'b', 'c']]


    class Foo(BaseModel):
        nums:List[int]
        bars: List[Bar] = []

        def resolve_bars(self):
            return [{'num': n} for n in self.nums]


    foo = Foo(nums=[1,2,3])
    with pytest.raises(TypeError) as e:
        await Resolver().resolve(foo)
    assert 'is not dict' in str(e.value)


@pytest.mark.asyncio
async def test_case_2():
    class Kar(BaseModel):
        name: str

        desc: str = ''
        def resolve_desc(self, ancestor_context):
            return f"{self.name} - {ancestor_context['bar_num']}"


    class Bar(BaseModel):
        __pydantic_resolve_expose__ = {'xnum': 'bar_num'}

        num: int

        kars: List[Kar] = []
        def resolve_kars(self):
            return [{'name': n} for n in ['a', 'b', 'c']]


    class Foo(BaseModel):
        nums:List[int]
        bars: List[Bar] = []

        def resolve_bars(self):
            return [{'num': n} for n in self.nums]


    foo = Foo(nums=[1,2,3])
    with pytest.raises(AttributeError) as e:
        await Resolver().resolve(foo)
    assert 'does not existed' in str(e.value)


# this component is stil in development
# the goal is to insert meta info into all pydantic models
# so that client component can access those info
# but for nested list, these may wasteful to insert meta info for each item

from pydantic import BaseModel, Field
from pydantic._internal._model_construction import ModelMetaclass


class MetaInfo(ModelMetaclass):
    def __new__(
        mcls,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, object],
        **kwargs,
    ) -> type:
        annotations = dict(namespace.get("__annotations__", {}))

        def _ensure_field(field_name: str, default_value: str) -> None:
            if field_name in annotations:
                return
            annotations[field_name] = str
            namespace[field_name] = Field(default=default_value, frozen=True)

        module_name = namespace.get("__module__", "")
        _ensure_field("$name", name)
        _ensure_field("$module", module_name)
        namespace["__annotations__"] = annotations

        return super().__new__(mcls, name, bases, namespace, **kwargs)


class Base(BaseModel, metaclass=MetaInfo):
    ...

if __name__ == "__main__":
    class A(BaseModel, metaclass=MetaInfo):
        id: int
    
    a = A(id=123)
    a.id = 222
    print(a)


from typing import Type, Union, List, Annotated
try:  # Python 3.10+ provides PEP 604 unions using types.UnionType
    from types import UnionType as _UnionType
except ImportError:  # pragma: no cover - prior to 3.10
    _UnionType = ()  # sentinel so membership tests still work

# Python <3.12 compatibility: TypeAliasType exists only from 3.12 (PEP 695)
try:  # pragma: no cover - import guard
    from typing import TypeAliasType  # type: ignore
except Exception:  # pragma: no cover
    class _DummyTypeAliasType:  # minimal sentinel so isinstance checks are safe
        pass
    TypeAliasType = _DummyTypeAliasType  # type: ignore

from typing import get_origin, get_args


def _is_optional(annotation):
    origin = get_origin(annotation)
    args = get_args(annotation)
    if origin is Union and type(None) in args:
        return True
    return False


def _is_list(annotation):
    return get_origin(annotation) in (list, List)


def shelling_type(tp):
    while _is_optional(tp) or _is_list(tp):
        tp = tp.__args__[0]
    return tp


def get_core_types(tp):
    """
    - get the core type
    - always return a tuple of core types
    """
    # Helpers
    def _unwrap_alias(t):
        """Unwrap PEP 695 type aliases by following __value__ repeatedly."""
        while isinstance(t, TypeAliasType) or (
            t.__class__.__name__ == 'TypeAliasType' and hasattr(t, '__value__')
        ):
            try:
                t = t.__value__
            except Exception:  # pragma: no cover - defensive
                break
        return t

    def _enqueue(items, q):
        for it in items:
            if it is not type(None):  # skip None in unions
                q.append(it)

    # Queue-based shelling to reach concrete core types
    queue: list[object] = [tp]
    result: list[object] = []

    while queue:
        cur = queue.pop(0)
        if cur is type(None):
            continue

        cur = _unwrap_alias(cur)

        # Handle Annotated[T, ...] as a shell
        if get_origin(cur) is Annotated:
            args = get_args(cur)
            if args:
                queue.append(args[0])
            continue

        # Handle Union / Optional / PEP 604 UnionType
        orig = get_origin(cur)
        if orig in (Union, _UnionType):
            args = get_args(cur)
            # push all non-None members back for further shelling
            _enqueue(args, queue)
            continue

        # Handle list shells
        if _is_list(cur):
            args = getattr(cur, "__args__", ())
            if args:
                queue.append(args[0])
            continue

        # If still an alias-like wrapper, unwrap again and re-process
        _cur2 = _unwrap_alias(cur)
        if _cur2 is not cur:
            queue.append(_cur2)
            continue

        # Otherwise treat as a concrete core type (could be a class, typing.Final, etc.)
        result.append(cur)

    return tuple(result)


def get_class_field_annotations(cls: Type):
    anno = cls.__dict__.get('__annotations__') or {}
    return anno.keys()


def get_type(v):
    return v.annotation





import inspect
from typing import Any, Callable, Optional
import pytest

# verify the Feasibility of implement depends 
class Loader:
    def load(self):
        print('load')

def Depend(  # noqa: N802
    dependency: Optional[Callable[..., Any]] = None 
) -> Any:
    return Depends(dependency=dependency)

class Depends:
    def __init__(
        self, dependency: Optional[Callable[..., Any]] = None
    ):
        self.dependency = dependency

class LoaderA(Loader):
    def load(self):
        print('load-a')

class TestClass:
    def resolve(self, loader = Depend(LoaderA)):
        loader.load()

def runner_maker():
    cache = {}
    counter = {
        "init_count": 0
    }

    def exec(t_method):
        signature = inspect.signature(t_method)
        params = {}

        for k, v in signature.parameters.items():
            if v.default == inspect._empty:
                continue

            if isinstance(v.default, Depends):
                cache_key = v.default.dependency.__name__
                hit = cache.get(cache_key)
                if hit:
                    instance = hit
                else:
                    instance = v.default.dependency()
                    cache[cache_key] = instance
                    counter["init_count"] += 1
                params[k] = instance
        t_method(**params)
        return counter
    return exec

@pytest.mark.asyncio
async def test_depend():
    run = runner_maker()
    t = TestClass()
    t2 = TestClass()
    t3 = TestClass()

    run(t.resolve)   # missing
    run(t2.resolve)  # hit
    counter = run(t3.resolve)  # hit

    assert counter["init_count"] == 1


from __future__ import annotations
from typing import List
import pytest
from collections import Counter
from aiodataloader import DataLoader
from pydantic import ConfigDict, BaseModel
from sqlalchemy import select
from sqlalchemy.ext.asyncio import async_sessionmaker
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from pydantic_resolve import Resolver, LoaderDepend, mapper, build_list

class Base(DeclarativeBase):
    pass

class Task(Base):
    __tablename__ = "task"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str]

class Comment(Base):
    __tablename__ = "comment"
    id: Mapped[int] = mapped_column(primary_key=True)
    task_id: Mapped[int] = mapped_column()
    content: Mapped[str]

class Feedback(Base):
    __tablename__ = "feedback"
    id: Mapped[int] = mapped_column(primary_key=True)
    comment_id: Mapped[int] = mapped_column()
    content: Mapped[str]

@pytest.mark.asyncio
async def test_sqlite_and_dataloader():
    counter = Counter()
    engine = create_async_engine(
        "sqlite+aiosqlite://",
        echo=False,
    )
    async_session = async_sessionmaker(engine, expire_on_commit=False)

    async def insert_objects() -> None:
        async with async_session() as session:
            async with session.begin():
                session.add_all(
                    [
                        Task(id=1, name="task-1"),
                        Task(id=2, name="task-2"),
                        Task(id=3, name="task-3"),

                        Comment(id=1, task_id=1, content="comment-1 for task 1"),
                        Comment(id=2, task_id=1, content="comment-2 for task 1"),
                        Comment(id=3, task_id=2, content="comment-1 for task 2"),

                        Feedback(id=1, comment_id=1, content="feedback-1 for comment-1"),
                        Feedback(id=2, comment_id=1, content="feedback-1 for comment-1"),
                        Feedback(id=3, comment_id=1, content="feedback-1 for comment-1"),
                    ]
                )

    class FeedbackLoader(DataLoader):
        async def batch_load_fn(self, comment_ids):
            counter['load-feedback'] += 1
            async with async_session() as session:
                res = await session.execute(select(Feedback).where(Feedback.comment_id.in_(comment_ids)))
                rows = res.scalars().all()
                return build_list(rows, comment_ids, lambda x: x.comment_id)

    class CommentLoader(DataLoader):
        async def batch_load_fn(self, task_ids):
            counter['load-comment'] += 1
            async with async_session() as session:
                res = await session.execute(select(Comment).where(Comment.task_id.in_(task_ids)))
                rows = res.scalars().all()
                return build_list(rows, task_ids, lambda x: x.task_id)

    class FeedbackSchema(BaseModel):
        id: int
        comment_id: int
        content: str
        model_config = ConfigDict(from_attributes=True)

    class CommentSchema(BaseModel):
        id: int
        task_id: int
        content: str
        feedbacks: List[FeedbackSchema]  = []
        @mapper(FeedbackSchema)
        def resolve_feedbacks(self, loader=LoaderDepend(FeedbackLoader)):
            return loader.load(self.id)
        model_config = ConfigDict(from_attributes=True)

    class TaskSchema(BaseModel):
        id: int
        name: str
        comments: List[CommentSchema]  = []
        @mapper(CommentSchema)
        def resolve_comments(self, loader=LoaderDepend(CommentLoader)):
            return loader.load(self.id)
        model_config = ConfigDict(from_attributes=True)

    async def init():
        async with engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)

    async def query():
        async with async_session() as session:
            tasks = (await session.execute(select(Task))).scalars().all()
            task_objs = [TaskSchema.model_validate(t) for t in tasks]
            resolved_results = await Resolver().resolve(task_objs)
            to_dict_arr = [r.model_dump() for r in resolved_results]
            return to_dict_arr

    await init()
    await insert_objects()
    result = await query()
    expected = [
        {
            'comments': [{'content': 'comment-1 for task 1',
            'feedbacks': [{'comment_id': 1,
                            'content': 'feedback-1 for comment-1',
                            'id': 1},
                            {'comment_id': 1,
                            'content': 'feedback-1 for comment-1',
                            'id': 2},
                            {'comment_id': 1,
                            'content': 'feedback-1 for comment-1',
                            'id': 3}],
            'id': 1,
            'task_id': 1},
                        {'content': 'comment-2 for task 1',
                            'feedbacks': [],
                            'id': 2,
                            'task_id': 1}],
            'id': 1,
            'name': 'task-1'},
            {'comments': [{'content': 'comment-1 for task 2',
                            'feedbacks': [],
                            'id': 3,
                            'task_id': 2}],
            'id': 2,
            'name': 'task-2'},
            {'comments': [], 'id': 3, 'name': 'task-3'}]

    assert result == expected
    assert counter['load-comment'] == 1  # batch_load_fn only called once
    assert counter['load-feedback'] == 1
    
    await query()

    assert counter['load-comment'] == 2  # Resolver + LoaderDepend can fix cache issue
    assert counter['load-feedback'] == 2


from __future__ import annotations
from typing import Tuple
from pydantic import BaseModel
from pydantic_resolve.analysis import Analytic

async def loader_fn(keys):
    return keys

class Student(BaseModel):
    zones: Tuple[int, int] = (0, 0)

def test_get_all_fields():
    # https://github.com/allmonday/pydantic2-resolve/issues/7
    result = Analytic().scan(Student)
    expect = {
        'test_specific_type.Student': {
            'resolve': [],
            'post': [],
            'object_fields': [],
            'expose_dict': {},
            'collect_dict': {}
            # ... others
        },
    }
    for k, v in result.items():
        assert expect[k].items() <= v.items()



from __future__ import annotations
from pydantic import BaseModel
from typing import List
from pydantic_resolve import Resolver, Collector
import pytest


class Root(BaseModel):
    list_a: List[A] = []
    def resolve_list_a(self):
        return data
    
    names: List[str] = []
    def post_names(self, collector=Collector('b_name')):
        return collector.values()

class A(BaseModel):
    list_b: List[B]

    names: List[str] = []
    def post_names(self, collector=Collector('b_name')):
        return collector.values()

class B(BaseModel):
    __pydantic_resolve_collect__ = {'name': 'b_name'}
    name: str


data = [
        {'list_b': [
            {'name': 'b1'},
            {'name': 'b2'},
        ]},
        {'list_b': [
            {'name': 'b3'},
            {'name': 'b4'},
        ]},
    ]

@pytest.mark.asyncio
async def test_level():
    r = Root()
    resolver = Resolver()
    r = await resolver.resolve(r)
    # print(resolver.object_collect_alias_map_store)
    assert r.model_dump() == {
        "list_a": [
            {
            "list_b": [
                {
                "name": "b1"
                },
                {
                "name": "b2"
                }
            ],
            "names": [
                "b1",
                "b2"
            ]
            },
            {
            "list_b": [
                {
                "name": "b3"
                },
                {
                "name": "b4"
                }
            ],
            "names": [
                "b3",
                "b4"
            ]
            }
        ],
        "names": [
            "b1",
            "b2",
            "b3",
            "b4"
        ]
    }


from __future__ import annotations
from typing import List, Optional
import asyncio
from pydantic import BaseModel
from pydantic_resolve import Resolver
import pytest

async def set_after(fut, value):
    await asyncio.sleep(.1)
    fut.set_result(value)

class Student(BaseModel):
    scores: List[int] = []
    async def resolve_scores(self) -> List[int]:
        return [1,2,3]

    age: Optional[int] = None
    def resolve_age(self) -> Optional[int]:
        return 12

    name: Optional[str] = None
    def resolve_name(self) -> Optional[str]:
        return 'name'
    
    future: Optional[str] = None
    def resolve_future(self) -> Optional[str]:
        loop = asyncio.get_running_loop()
        fut = loop.create_future()
        loop.create_task(set_after(fut, 'hello'))
        return fut

@pytest.mark.asyncio
async def test_resolve_future():
    stu = Student()
    result = await Resolver().resolve(stu)
    expected = {
        'name': 'name',
        'scores': [1,2,3],
        'age': 12,
        'future': 'hello'
    }
    assert result.model_dump() == expected



from pydantic import BaseModel
from pydantic_resolve.analysis import Analytic

class E(BaseModel):
    id: int
    def resolve_id(self):
        return self.id

class D(BaseModel):
    id: int
    e: E

class C(BaseModel):
    """
    INTRODUCTION:

    before ver: 1.11.8
    d, or e will be removed from the object_fields
    it is due to the _has_config function do not verify the len of 
    object_fields.

    but in fact, if a kls is cached, it should not take party in judging
    `should_treverse`, so from 1.11.9 it will return None
    """

    id: int
    d: D
    e: D

class BaseB(BaseModel):
    c: C

class B(BaseB):
    id: int

class A(BaseModel):
    id: int
    c: E  # travers once, then populate the cache info
    b: B


def test_long_distance_resolve():
    result = Analytic().scan(A)
    prefix = 'tests.analysis.test_analysis_object_fields_4'
    expect = {
        f'{prefix}.A': {
            'resolve': [],
            'post': [],
            'object_fields': [ 'c', 'b'],
            'expose_dict': {},
            'collect_dict': {},
        },
        f'{prefix}.B': {
            'resolve': [],
            'post': [],
            'object_fields': ['c'],
            'expose_dict': {},
            'collect_dict': {},
        },
        f'{prefix}.C': {
            'resolve': [],
            'post': [],
            'object_fields': ['d', 'e'],
            'expose_dict': {},
            'collect_dict': {},
        },

        # before and include 1.11.8, it looks like
        # f'{prefix}.C': {
        #     'resolve': [],
        #     'post': [],
        #     'object_fields': ['d'],  # or ['e']
        #     'expose_dict': {},
        #     'collect_dict': {},
        # },

        f'{prefix}.D': {
            'resolve': [],
            'post': [],
            'object_fields': ['e'],
            'expose_dict': {},
            'collect_dict': {},
        },
        f'{prefix}.E': {
            'resolve': ['resolve_id'],
            'post': [],
            'object_fields': [],
            'expose_dict': {},
            'collect_dict': {},
        }
    }
    from pprint import pprint
    pprint(result)
    for k, v in result.items():
        assert expect[k].items() <= v.items()


from typing import Type, Dict, Any, List, Tuple, Set, Literal
from pydantic import BaseModel, create_model, model_validator
import copy
import pydantic_resolve.constant as const
from pydantic_resolve.utils.expose import ExposeAs
from pydantic_resolve.utils.collector import SendTo


class SubsetConfig(BaseModel): 
    kls: Type[BaseModel]  # parent class

    # fields and omit_fields are exclusive
    fields: List[str] | Literal["all"] | None = None
    omit_fields: List[str] | None = None

    # set Field(exclude=True) for these fields
    excluded_fields: List[str] | None = None

    expose_as: List[Tuple[str, str]] | None = None
    send_to: List[Tuple[str, Tuple[str, ...] | str]] | None = None

    @model_validator(mode='after')
    def validate_fields(self):
        if self.fields is not None and self.omit_fields is not None:
            raise ValueError("fields and omit_fields are exclusive")
        
        if self.fields is None and self.omit_fields is None:
            raise ValueError("fields or omit_fields must be provided")
        return self


def _extract_field_infos(kls: Type[BaseModel], field_names: List[str]) -> Dict[str, Any]:
    field_definitions = {}
    
    for field_name in field_names:
        field = kls.model_fields.get(field_name)
        if not field:
            raise AttributeError(f'field "{field_name}" not existed in {kls.__name__}')
        
        field_definitions[field_name] = (field.annotation, field)
    
    return field_definitions


def _get_kls_config(kls: Type[BaseModel]) -> Any:
    return getattr(kls, 'model_config', None)


def _get_kls_validators(kls: Type[BaseModel], field_names: List[str]) -> Dict[str, Any]:

    validators = {}
    decorators = getattr(kls, '__pydantic_decorators__', None)

    if decorators and hasattr(decorators, 'field_validators'):
        for validator_name, decorator_info in decorators.field_validators.items():
            # example: @field_validator('email', 'name') -> decorator_info.info.fields = ['email', 'name']
            if any(field in field_names for field in decorator_info.info.fields):
                validators[validator_name] = getattr(kls, validator_name)
    
    return validators


def _apply_validators(model_class: Type[BaseModel], validators: Dict[str, Any]) -> None:
    for name, validator in validators.items():
        setattr(model_class, name, validator)


def _extract_extra_fields_from_namespace(namespace: Dict[str, Any], disallow: Set[str]) -> Dict[str, Tuple[Any, Any]]:
    """Extract extra field definitions declared on the subset class body.

    Only annotated attributes are considered as Pydantic fields. The result is a
    dict mapping field name -> (annotation, default | Ellipsis) suitable for create_model.

    Args:
        namespace: The class namespace passed into the metaclass with attributes and annotations.
        disallow: Field names that are already selected from the parent and thus cannot be redefined here.

    Raises:
        ValueError: If an extra field duplicates a subset field name.
    """
    annotations: Dict[str, Any] = namespace.get('__annotations__', {}) or {}
    extras: Dict[str, Tuple[Any, Any]] = {}

    for fname, anno in annotations.items():
        if fname in (const.ENSURE_SUBSET_DEFINITION, const.ENSURE_SUBSET_DEFINITION_SHORT):
            continue
        if fname in disallow:
            raise ValueError(f'additional field "{fname}" duplicates subset field')
        default = namespace.get(fname, ...)
        extras[fname] = (anno, default)

    return extras


def _validate_subset_fields(fields: Any):
    if not isinstance(fields, (list, tuple)):
        raise TypeError('fields must be a list or tuple of field names')

    hit = set()
    for f in fields:
        if not isinstance(f, str):
            raise TypeError('each field name must be a string')
        else:
            if f in hit:
                raise ValueError(f'duplicate field name "{f}" in subset fields')
            hit.add(f)


def create_subset(parent: Type[BaseModel], fields: List[str], name: str = "SubsetModel") -> Type[BaseModel]:
    """
    Create a subset model from a parent BaseModel using Pydantic's create_model.
    
    Args:
        parent: Parent BaseModel class to create subset from
        fields: List of field names to include in subset
        name: Name of the new subset class (default: "SubsetModel")
        
    Returns:
        A new BaseModel class containing only the specified fields
    """
    if not issubclass(parent, BaseModel):
        raise TypeError('parent must be a pydantic BaseModel')
    
    _validate_subset_fields(fields)
    field_infos = _extract_field_infos(parent, fields)
    validators = _get_kls_validators(parent, fields)
    create_model_kwargs = {}

    config = _get_kls_config(parent)
    if config:
        create_model_kwargs['__config__'] = config
    
    subset_class = create_model(name, **field_infos, **create_model_kwargs)

    _apply_validators(subset_class, validators)
    setattr(subset_class, const.ENSURE_SUBSET_REFERENCE, parent)

    return subset_class


class SubsetMeta(type):
    def __new__(cls, name, bases, namespace, **kwargs):
        # subset_info is expected to be a tuple of (parent_class, list_of_field_names) or SubsetConfig
        subset_info = namespace.get(const.ENSURE_SUBSET_DEFINITION) or namespace.get(const.ENSURE_SUBSET_DEFINITION_SHORT)
            
        # Allow defining the base marker class without warning, bypass
        if name == 'DefineSubset' and not bases:
            return super().__new__(cls, name, bases, namespace, **kwargs)

        if not subset_info:
            raise ValueError(f'Class {name} must define {const.ENSURE_SUBSET_DEFINITION} to use SubsetMeta')

        if isinstance(subset_info, SubsetConfig):
            parent_kls = subset_info.kls

            if subset_info.fields is not None:
                if subset_info.fields == "all":  # special value to include all fields
                    subset_fields = list(parent_kls.model_fields.keys())
                else:
                    subset_fields = subset_info.fields
            else:
                all_fields = list(parent_kls.model_fields.keys())
                omit_set = set(subset_info.omit_fields)
                subset_fields = [f for f in all_fields if f not in omit_set]
        else:
            parent_kls, subset_fields = subset_info

        _validate_subset_fields(subset_fields)
        field_infos = _extract_field_infos(parent_kls, subset_fields)

        if isinstance(subset_info, SubsetConfig):
            if subset_info.excluded_fields:
                for field_name in subset_info.excluded_fields:
                    if field_name in field_infos:
                        annotation, field_info = field_infos[field_name]
                        new_field_info = copy.deepcopy(field_info)
                        new_field_info.exclude = True
                        field_infos[field_name] = (annotation, new_field_info)

            if subset_info.expose_as:
                for field_name, alias in subset_info.expose_as:
                    if field_name in field_infos:
                        annotation, field_info = field_infos[field_name]
                        new_field_info = copy.deepcopy(field_info)
                        new_field_info.metadata.append(ExposeAs(alias))
                        field_infos[field_name] = (annotation, new_field_info)
            
            if subset_info.send_to:
                for field_name, target in subset_info.send_to:
                    if field_name in field_infos:
                        annotation, field_info = field_infos[field_name]
                        new_field_info = copy.deepcopy(field_info)
                        new_field_info.metadata.append(SendTo(target))
                        field_infos[field_name] = (annotation, new_field_info)

        # Then extract extra fields defined in class body
        extra_fields = _extract_extra_fields_from_namespace(namespace, set(subset_fields))

        # Parent validators and config should still apply to subset fields
        create_model_kwargs: Dict[str, Any] = {}
        attributes_to_attach: Dict[str, Any] = {}
        methods_to_attach: Dict[str, Any] = {}

        validators = _get_kls_validators(parent_kls, subset_fields)
        config = _get_kls_config(parent_kls)
        if config:
            create_model_kwargs['__config__'] = config

        # Use the caller's module for better reprs and pickling
        create_model_kwargs['__module__'] = namespace.get('__module__', __name__)

        for ns_key, ns_value in namespace.items():
            # definition like __pydatnic_resolve_expose__ or __pydantic_resolve_collect__ will be kept, just in case.
            # except __pydantic_resolve_subset__  ( __subset__ will also be ignored)
            if ns_key.startswith('__pydantic_resolve') and ns_key != const.ENSURE_SUBSET_DEFINITION:
                attributes_to_attach[ns_key] = ns_value
                continue

            if callable(ns_value) and (ns_key.startswith(const.RESOLVE_PREFIX) or ns_key.startswith(const.POST_PREFIX)):
                methods_to_attach[ns_key] = ns_value

        field_definitions = {}
        field_definitions.update(field_infos)
        field_definitions.update(extra_fields)

        subset_class = create_model(
            name,
            **field_definitions,
            **create_model_kwargs
        )

        _apply_validators(subset_class, validators)

        for method_name, method in methods_to_attach.items():
            setattr(subset_class, method_name, method)
        for attr_name, attr_value in attributes_to_attach.items():
            setattr(subset_class, attr_name, attr_value)
        setattr(subset_class, const.ENSURE_SUBSET_REFERENCE, parent_kls)

        return subset_class


class DefineSubset(metaclass=SubsetMeta):
    pass



from typing import List, Optional
from pydantic import BaseModel
from pydantic_resolve import Resolver, mapper, LoaderDepend
import pytest

# define loader functions
async def friends_batch_load_fn(names):
    mock_db = {
        'tangkikodo': ['tom', 'jerry'],
        'john': ['mike', 'wallace'],
        'trump': ['sam', 'jim'],
        'sally': ['sindy', 'lydia'],
    }
    return [mock_db.get(name, []) for name in names]

async def cash_batch_load_fn(names):
    mock_db = {
        'jerry':200, 'mike': 3000, 'wallace': 400, 'sam': 500,
        'jim': 600, 'sindy': 700, 'lydia': 800, 'tangkikodo': 900, 'john': 1000,
        'trump': 1200
    }
    result = []
    for name in names:
        n = mock_db.get(name, None)
        result.append({'number': n} if n else None)  # conver to auto mapping compatible style
    return result

# define schemas
class Cash(BaseModel):
    number: Optional[int] = None

class Friend(BaseModel):
    name: str

    cash: Optional[Cash] = None
    @mapper(Cash)  # auto mapping
    def resolve_cash(self, contact_loader=LoaderDepend(cash_batch_load_fn)):
        return contact_loader.load(self.name)
    
    has_cash: bool = False
    def post_has_cash(self):
        # self.has_cash = self.cash is not None
        return self.cash is not None

    

class User(BaseModel):
    name: str
    age: int

    friends: List[Friend] = []
    @mapper(lambda names: [Friend(name=name) for name in names])
    def resolve_friends(self, friend_loader=LoaderDepend(friends_batch_load_fn)):
        return friend_loader.load(self.name)
    
    has_cash: bool = False
    def post_has_cash(self):
        return any([f.has_cash for f in self.friends])

class Root(BaseModel):
    users: List[User] = []
    @mapper(lambda items: [User(**item) for item in items])
    def resolve_users(self):
        return [
            {"name": "tangkikodo", "age": 19}, 
            {"name": "noone", "age": 19}, 
        ]
    hello: str = ''
    def post_default_handler(self, context):
        self.hello = f'hello, {context["world"]}'

@pytest.mark.asyncio
async def test_post_methods():
    root = Root()
    root = await Resolver(context={"world": "new world"}).resolve(root)
    dct = root.model_dump()
    assert dct == {'users': [{'age': 19,
                     'has_cash': True,
                     'friends': [{'has_cash': False,'cash': None, 'name': 'tom'},
                                 {'has_cash': True, 'cash': {'number': 200}, 'name': 'jerry'}],
                     'name': 'tangkikodo'},
                     {'name': 'noone', 'age': 19, 'friends': [], 'has_cash':False }
                     ],
                    'hello': 'hello, new world'}



from aiodataloader import DataLoader
from pydantic import BaseModel

BOOKS = {
    1: [{'name': 'book1'}, {'name': 'book2'}],
    2: [{'name': 'book3'}, {'name': 'book4'}],
    3: [{'name': 'book1'}, {'name': 'book2'}],
}

class Book(BaseModel):
    name: str
    public: str = 'public'

class BookLoader(DataLoader):
    async def batch_load_fn(self, keys):
        books = [[Book(**bb) for bb in BOOKS.get(k, [])] for k in keys]
        return books


import copy
import inspect
from typing import List, Type, Dict, Optional, Tuple, TypedDict
from inspect import isfunction
from collections import defaultdict
from pydantic import BaseModel
import pydantic_resolve.constant as const
import pydantic_resolve.utils.class_util as class_util
from pydantic_resolve.utils.collector import ICollector, pre_generate_collector_config
from pydantic_resolve.utils.depend import Depends
from pydantic_resolve.utils.er_diagram import ErLoaderPreGenerator
from pydantic_resolve.exceptions import ResolverTargetAttrNotFound, MissingCollector
from pydantic_resolve.utils.expose import pre_generate_expose_config

class DataLoaderType(TypedDict):
    param: str
    kls: Type
    path: str
    request_type: List[Type]

class CollectorType(TypedDict):
    field: str 
    param: str
    instance: object
    alias: str
    
class ResolveMethodType(TypedDict):
    trim_field: str
    context: bool
    parent: bool
    ancestor_context: bool
    dataloaders: List[DataLoaderType]

class PostMethodType(TypedDict):
    trim_field: str
    context: bool
    parent: bool
    ancestor_context: bool
    dataloaders: List[DataLoaderType]
    collectors: List[CollectorType] 

class PostDefaultHandlerType(TypedDict):
    context: bool
    parent: bool
    ancestor_context: bool
    collectors: List[CollectorType]

class KlsMetaType(TypedDict):
    resolve: List[str]
    post: List[str]
    resolve_params: Dict[str, ResolveMethodType]
    post_params: Dict[str, PostMethodType]
    post_default_handler_params: Optional[PostDefaultHandlerType]
    raw_object_fields: List[str] # store the original field names
    object_fields: List[str]
    expose_dict: dict
    collect_dict: dict
    kls: Type
    has_context: bool
    should_traverse: bool

class LoaderQueryMetaRequestType(TypedDict):
    name: Type
    fields: List[str]

class LoaderQueryMeta(TypedDict):
    required_types: List
    fields: List[str]

MetaType = Dict[str, KlsMetaType]

class MappedMetaMemberType(KlsMetaType):
    kls_path: str
    alias_map_proto: Dict[str, Dict[Tuple[str, str, str], object]]

MappedMetaType = Dict[Type, MappedMetaMemberType]


def _scan_resolve_method(method, field: str, request_types: List[Type]) -> ResolveMethodType:
    result: ResolveMethodType = {
        'trim_field': field.replace(const.RESOLVE_PREFIX, ''),
        'context': False,
        'parent': False,
        'ancestor_context': False,
        'dataloaders': []  # collect func or class, do not create instance
    }
    signature = inspect.signature(method)

    if signature.parameters.get('context'):
        result['context'] = True

    if signature.parameters.get('ancestor_context'):
        result['ancestor_context'] = True

    if signature.parameters.get('parent'):
        result['parent'] = True

    for name, param in signature.parameters.items():
        if isinstance(param.default, Depends):
            info: DataLoaderType = { 
                'param': name,
                'kls': param.default.dependency,  # for later initialization
                'path': class_util.get_kls_full_name(param.default.dependency),
                'request_type': request_types
            }
            result['dataloaders'].append(info)

    for name, param in signature.parameters.items():
        if isinstance(param.default, ICollector):
            raise AttributeError("Collector is not available in resolve_method")

    return result


def _scan_post_method(method, field: str, request_types: List[Type]) -> PostMethodType:
    result: PostMethodType = {
        'trim_field': field.replace(const.POST_PREFIX, ''),
        'context': False,
        'ancestor_context': False,
        'parent': False,
        'dataloaders': [],
        'collectors': []
    }
    signature = inspect.signature(method)

    if signature.parameters.get('context'):
        result['context'] = True

    if signature.parameters.get('ancestor_context'):
        result['ancestor_context'] = True

    if signature.parameters.get('parent'):
        result['parent'] = True

    
    for name, param in signature.parameters.items():
        if isinstance(param.default, Depends):
            loader_info: DataLoaderType = { 
                'param': name,
                'kls': param.default.dependency,  # for later initialization
                'path': class_util.get_kls_full_name(param.default.dependency),
                'request_type': request_types
            }
            result['dataloaders'].append(loader_info)

        if isinstance(param.default, ICollector):
            info: CollectorType = {
                'field': field,
                'param': name,
                'instance': param.default,
                'alias': param.default.alias 
            }
            result['collectors'].append(info)

            
    return result


def _scan_post_default_handler(method) -> PostDefaultHandlerType:
    result: PostDefaultHandlerType = {
        'context': False,
        'ancestor_context': False,
        'parent': False,
        'collectors': []
    }
    signature = inspect.signature(method)

    if signature.parameters.get('context'):
        result['context'] = True

    if signature.parameters.get('ancestor_context'):
        result['ancestor_context'] = True

    if signature.parameters.get('parent'):
        result['parent'] = True

    for name, param in signature.parameters.items():
        if isinstance(param.default, ICollector):
            info: CollectorType = {
                'field': const.POST_DEFAULT_HANDLER,
                'param': name,
                'instance': param.default,
                'alias': param.default.alias 
            }
            result['collectors'].append(info)

    return result


class Analytic:
    """
    Analyze pydantic models to extract metadata for resolution.
    
    This class traverses the pydantic model tree to collect:
    - resolve methods
    - post methods
    - expose configurations
    - collector configurations
    - object fields that need traversal
    """
    def __init__(self, er_pre_generator: Optional[ErLoaderPreGenerator]=None) -> None:
        self.expose_set = set()
        self.collect_set = set()
        self.metadata: MetaType = {}
        self.er_pre_generator = er_pre_generator

    def _get_request_type_for_loader(self, object_field_pairs, field_name: str):
        return object_field_pairs.get(field_name)

    def _get_all_fields_and_object_fields(self, kls: Type):
        if class_util.safe_issubclass(kls, BaseModel):
            all_fields = set(class_util.get_pydantic_field_keys(kls))
            object_fields = list(class_util.get_pydantic_fields(kls))  # dive and recursively analysis
        else:
            raise TypeError('invalid type: should be pydantic object')  # noqa
        return all_fields, object_fields, {k: v for k, v in object_fields}

    def _has_post_default_handler(self, kls: Type) -> bool:
        fields = dir(kls)
        fields = [f for f in fields if f == const.POST_DEFAULT_HANDLER and isfunction(getattr(kls, f))]
        return len(fields) > 0

    def _get_resolve_and_post_fields(self, kls: Type):
        fields = dir(kls)

        resolve_fields = [f for f in fields if f.startswith(const.RESOLVE_PREFIX) and isfunction(getattr(kls, f))]
        post_fields = [
            f for f in fields if f.startswith(const.POST_PREFIX)
            and f != const.POST_DEFAULT_HANDLER
            and isfunction(getattr(kls, f))
        ]
        fields_with_resolver = {field.replace(const.RESOLVE_PREFIX, '') for field in resolve_fields}
        return resolve_fields, post_fields, fields_with_resolver

    def _validate_resolve_and_post_fields(self, resolve_fields, post_fields, all_fields):
        for field in resolve_fields:
            resolve_field = field.replace(const.RESOLVE_PREFIX, '')
            if resolve_field not in all_fields:
                raise ResolverTargetAttrNotFound(f"attribute {resolve_field} not found")

        for field in post_fields:
            post_field = field.replace(const.POST_PREFIX, '')
            if post_field not in all_fields:
                raise ResolverTargetAttrNotFound(f"attribute {post_field} not found")

    def _validate_expose(self, expose_dict, kls_name: str):
        if not isinstance(expose_dict, dict):
            raise TypeError(f'{const.EXPOSE_TO_DESCENDANT} is not dict')
        for _, v in expose_dict.items():
            if v in self.expose_set:
                raise ValueError(f'Expose alias name conflicts, please check: {kls_name}')
            self.expose_set.add(v)

    def _add_collector_info(self, post_params):
        """
        add collector alias during _walker (breadth first traversal)
        so that it should ensure the definition of collector is always before usage
        which means collector should always be defined before ancestor nodes actually send data.
        """
        for _, param in post_params.items():
            for collector in param['collectors']:
                self.collect_set.add(collector['alias'])

    def _add_collector_info_for_default_handler(self, post_default_params):
        """
        same like _add_collector_info
        """
        if post_default_params is not None:
            for collector in post_default_params['collectors']:
                self.collect_set.add(collector['alias'])

    def _validate_collector(self, collect_dict, kls_name: str):
        """
        validate if the collector is declared in ancestor nodes.

        available format of `__pydantic_resolve_collect__`

        class A(BaseModel):
            __pydantic_resolve_collect__ = {
                'name': 'collector_a',                                  # send name to collector_a
                ('name', 'age'): 'collector_b',                         # send ('name', 'age') to collector_b
                'name': ('collector_c', 'collector_d'),                 # send name to both collector_c and d
                ('name', 'age'): ('collector_e', 'collector_f'),        # send ('name', 'age') to both collector_e and f
            }
        """
        if type(collect_dict) is not dict:
            raise AttributeError(f'{const.COLLECTOR_CONFIGURATION} is not dict')

        for _, collector in collect_dict.items():
            collectors = collector if isinstance(collector, (list, tuple)) else (collector,)
            for c in collectors:
                if c not in self.collect_set:
                    raise MissingCollector(f'Collector alias name not found in ancestor, please check: {kls_name}')

    def _has_config(self, info: KlsMetaType):
        result = (
            len(info['resolve']) > 0 or
            len(info['post']) > 0 or
            len(info['collect_dict']) > 0 or
            len(info['expose_dict']) > 0 or
            info['post_default_handler_params'] is not None
        )
        return result

    def _populate_ancestors(self, parents):
        """
        mark should traverse to plan a minimal traversal path
        """
        for field, kls_name in parents:
            # normally, array size is small
            if field in self.metadata[kls_name]['raw_object_fields'] and \
                    field not in self.metadata[kls_name]['object_fields']:
                self.metadata[kls_name]['object_fields'].append(field)
            self.metadata[kls_name]['should_traverse'] = True

    def _walker(self, kls: Type, ancestors: List[Tuple[str, str]]):
        kls_name = class_util.get_kls_full_name(kls)
        hit = self.metadata.get(kls_name)
        if hit:
            # if populated by previous node, or self has_config
            if hit['should_traverse'] or self._has_config(hit):
                self._populate_ancestors(ancestors)
            return

        # based on ER diagram, generate resolve + dataloader by field's metadata settings
        if self.er_pre_generator: 
            self.er_pre_generator.prepare(kls)
        
        # generate expose and collector config based on field's metadata settings
        # __pydantic_resolve_expose__ and __pydantic_resolve_collect__
        pre_generate_expose_config(kls)
        pre_generate_collector_config(kls)

        # prepare fields, with resolve_, post_ reserved
        all_fields, object_fields, object_field_pairs = self._get_all_fields_and_object_fields(kls)
        resolve_fields, post_fields, fields_with_resolver = self._get_resolve_and_post_fields(kls)
        self._validate_resolve_and_post_fields(resolve_fields, post_fields, all_fields)
        object_fields_without_resolver = [a[0] for a in object_fields if a[0] not in fields_with_resolver]

        # - scan expose and collect (__pydantic_resolve_xxx__)
        expose_dict = getattr(kls, const.EXPOSE_TO_DESCENDANT, {})
        collect_dict = getattr(kls, const.COLLECTOR_CONFIGURATION, {})
        self._validate_expose(expose_dict, kls_name)

        resolve_params = {
            field: _scan_resolve_method(
                getattr(kls, field),
                field,
                self._get_request_type_for_loader(
                    object_field_pairs,
                    field.replace(const.RESOLVE_PREFIX, '')))
            for field in resolve_fields
        }
        post_params = {
            field: _scan_post_method(
                getattr(kls, field),
                field,
                self._get_request_type_for_loader(
                    object_field_pairs,
                    field.replace(const.POST_PREFIX, '')))
            for field in post_fields
        }
        post_default_handler_params = _scan_post_default_handler(getattr(kls, const.POST_DEFAULT_HANDLER)) if self._has_post_default_handler(kls) else None

        # check context config
        resolve_context = any([p['context'] for p in resolve_params.values()])
        post_context = any([p['context'] for p in post_params.values()])
        post_default_context = post_default_handler_params['context'] if post_default_handler_params else False
        has_context = resolve_context or post_context or post_default_context

        # check collector
        self._add_collector_info(post_params)
        self._add_collector_info_for_default_handler(post_default_handler_params)
        self._validate_collector(collect_dict, kls_name)

        info: KlsMetaType = {
            'resolve': resolve_fields,
            'resolve_params': resolve_params,
            'post': post_fields,
            'post_params': post_params,
            'post_default_handler_params': post_default_handler_params,
            'raw_object_fields': object_fields_without_resolver,
            'object_fields': [],
            'expose_dict': expose_dict,
            'collect_dict': collect_dict,
            'kls': kls,
            'has_context': has_context,
            'should_traverse': False
        }
        self.metadata[kls_name] = info

        # visit fields (pydantic class) without resolve method
        for field, shelled_types in (obj for obj in object_fields if obj[0] in object_fields_without_resolver):
            for shelled_type in shelled_types:
                self._walker(shelled_type, ancestors + [(field, kls_name)])

        # visit fields with resolve method
        for field, shelled_types in (obj for obj in object_fields if obj[0] in fields_with_resolver):
            for shelled_type in shelled_types:
                self._walker(shelled_type, ancestors + [(field, kls_name)])

        should_traverse = info['should_traverse'] or self._has_config(info)

        if should_traverse:
            self._populate_ancestors(ancestors)

        info['should_traverse'] = should_traverse

    def scan(self, root_class: Type) -> MetaType:
        """Public method to perform metadata scan and return the metadata map."""
        # reset state for each scan
        self.expose_set = set()
        self.collect_set = set()
        self.metadata = {}

        core_types = class_util.get_core_types(root_class)

        for ct in core_types:
            class_util.update_forward_refs(ct)

        for ct in core_types:
            self._walker(ct, [])

        return self.metadata


def convert_metadata_key_as_kls(metadata: MetaType) -> MappedMetaType:
    """ use kls as map key for performance """
    kls_metadata = {}
    for k, v in metadata.items():
        _kls_meta: MappedMetaMemberType = {
            **v, 
            'kls_path':k,
            'alias_map_proto': {}
        }
        alias_map = _calc_alias_map_from_collectors(_kls_meta)
        _kls_meta['alias_map_proto'] = alias_map
        kls_metadata[v['kls']] = _kls_meta
    
    return kls_metadata


def _calc_alias_map_from_collectors(kls_meta: MappedMetaMemberType):
    post_params = kls_meta['post_params']
    kls_path = kls_meta['kls_path']

    alias_map = defaultdict(dict)

    def add_collector(collector: CollectorType):
        """
        class A(BaseModel):
            def post_name(collector=Collector('name')):
                ...
        kls_path: module_name.A
        field: post_name
        param: collector_a

        sign = ('module_name.A', 'post_name', 'collector_a')

        return { 'name': { sign: collector_instance } }
        """
        sign = get_collector_sign(kls_path, collector)
        alias_map[collector['alias']][sign] = collector['instance']

    for _, param in post_params.items():
        for collector in param['collectors']:
            add_collector(collector)

    # post_default_handler
    post_default_handler_params = kls_meta['post_default_handler_params']
    if post_default_handler_params:
        for collector in post_default_handler_params['collectors']:
            add_collector(collector)

    return alias_map


def is_acceptable_kls(kls: Type) -> bool:
    return class_util.safe_issubclass(kls, BaseModel)


def is_acceptable_instance(target: object):
    return isinstance(target, BaseModel)


def get_resolve_fields_and_object_fields_from_object(node: object, kls: Type, mapped_metadata: MappedMetaType):
    """
    resolve_fields: a field with resolve_field method
    object_fields: a field without resolve_field method but is an object (should traversal)
    """
    resolve_fields, object_fields = [], []
    kls_meta = mapped_metadata.get(kls)

    if kls_meta is None:
        return [], []

    for resolve_field in kls_meta['resolve']:
        attr = getattr(node, resolve_field)
        trim_field = kls_meta['resolve_params'][resolve_field]['trim_field']
        resolve_fields.append((resolve_field, trim_field, attr))

    for attr_name in kls_meta['object_fields']:
        attr = getattr(node, attr_name)
        object_fields.append((attr_name, attr))

    return resolve_fields, object_fields


def get_post_methods(node: object, kls: Type, mapped_metadata: MappedMetaType):
    kls_meta = mapped_metadata.get(kls)

    if kls_meta is None:
        return []

    for post_field in kls_meta['post']:
        attr = getattr(node, post_field)
        trim_field = kls_meta['post_params'][post_field]['trim_field']
        yield post_field, trim_field, attr


def get_resolve_method_param(kls: Type, resolve_field: str, mapped_metadata: MappedMetaType):
    kls_meta = mapped_metadata.get(kls, {})
    return kls_meta['resolve_params'][resolve_field]


def get_post_method_params(kls, post_field, mapped_metadata: MappedMetaType):
    kls_meta = mapped_metadata.get(kls, {})
    return kls_meta['post_params'][post_field]


def get_post_default_handler_params(kls, mapped_metadata: MappedMetaType):
    kls_meta = mapped_metadata.get(kls, {})
    return kls_meta['post_default_handler_params']


def get_collector_sign(kls_path: str, collector: CollectorType):
    return (kls_path, collector['field'], collector['param'])


def generate_alias_map_with_cloned_collector(kls: Type, mapped_metadata: MappedMetaType):
    kls_meta = mapped_metadata.get(kls, {})
    return { alias: {
        sign: copy.deepcopy(collector) for sign, collector in v.items()
    } for alias, v in kls_meta['alias_map_proto'].items()}


def get_collector_candidates(kls, metadata: MappedMetaType):
    kls_meta = metadata.get(kls, {})
    collect_dict = kls_meta['collect_dict']

    for field, alias in collect_dict.items():
        yield field, alias


def has_context(mapped_metadata: MappedMetaType):
    return any([ m['has_context'] for m in mapped_metadata.values()])


"""Tests for create_subset function in subset.py"""

import pytest
from typing import Optional, Annotated
from pydantic import BaseModel, Field
from pydantic_resolve.utils.subset import create_subset, DefineSubset
from pydantic_resolve import Resolver, Collector
import pydantic_resolve.constant as const
from pydantic_resolve.utils.subset import SubsetConfig


class TestCreateSubset:
    """Test cases for create_subset function."""
    
    def test_basic_subset_creation(self):
        """Test basic subset creation with required fields."""
        class Parent(BaseModel):
            id: int
            name: Annotated[str, 'tangkikodo']
            age: int
            email: str
            items: list[str]
        
        Subset = create_subset(Parent, ['id', 'name', 'items'], 'TestSubset')
        
        # Test that we can create an instance
        instance = Subset(id=1, name='test', items=['a', 'b', 'c'])
        assert instance.id == 1
        assert instance.name == 'test'
        assert instance.items == ['a', 'b', 'c']
        
        # Test that excluded fields are not present in the model
        field_names = list(Subset.model_fields.keys())

        assert len(Subset.model_fields.get('name').metadata) == 1
        assert Subset.model_fields.get('name').metadata[0] == 'tangkikodo'
        
        assert 'id' in field_names
        assert 'name' in field_names
        assert 'age' not in field_names
        assert 'email' not in field_names

        # Test that extra fields are ignored (Pydantic default behavior)
        instance_with_extra = Subset(id=2, name='test2', age=25, items=['x', 'y', 'z'])  # age is ignored
        assert instance_with_extra.id == 2
        assert instance_with_extra.name == 'test2'
        assert not hasattr(instance_with_extra, 'age')
    
    def test_subset_with_optional_fields(self):
        """Test subset creation with optional fields."""
        class Parent(BaseModel):
            id: int
            name: str
            description: Optional[str] = None
            active: bool = True
        
        Subset = create_subset(Parent, ['id', 'description', 'active'], 'OptionalSubset')
        
        # Test with default values
        instance1 = Subset(id=1)
        assert instance1.id == 1
        assert instance1.description is None
        assert instance1.active is True
        
        # Test with provided values
        instance2 = Subset(id=2, description='test desc', active=False)
        assert instance2.id == 2
        assert instance2.description == 'test desc'
        assert instance2.active is False
    
    def test_subset_with_field_constraints(self):
        """Test subset creation preserves field constraints."""
        class Parent(BaseModel):
            id: int = Field(gt=0, description="ID must be positive")
            name: str = Field(min_length=2, max_length=50)
            score: float = Field(ge=0.0, le=100.0)
        
        Subset = create_subset(Parent, ['id', 'name'], 'ConstrainedSubset')
        
        # Test valid values
        instance = Subset(id=1, name='test')
        assert instance.id == 1
        assert instance.name == 'test'
        
        # Test constraint validation (if constraints are preserved)
        # Note: This may depend on how well create_model preserves constraints
        try:
            Subset(id=0, name='a')  # id should be > 0, name too short
            # If no exception is raised, constraints might not be fully preserved
            # This is expected behavior with basic create_model usage
        except Exception:
            # If constraints are preserved, we expect validation errors
            pass
    
    def test_subset_with_validators(self):
        """Test subset creation with parent validators."""
        from pydantic import field_validator
        
        class Parent(BaseModel):
            id: int
            name: str
            email: str
            
            @field_validator('name', 'email')
            @classmethod
            def validate_name(cls, v):
                if len(v) < 2:
                    raise ValueError('Name must be at least 2 characters')
                return v

            @field_validator('email')
            @classmethod
            def validate_email(cls, v):
                if '@' not in v:
                    raise ValueError('Email must has @ symbol')
                return v
        
        Subset = create_subset(Parent, ['id', 'name'], 'ValidatorSubset')
        
        # Test that valid data works
        instance = Subset(id=1, name='test')
        assert instance.id == 1
        assert instance.name == 'test'
        
        # Test that the validator method is copied to the subset
        assert hasattr(Subset, 'validate_name')
        assert callable(getattr(Subset, 'validate_name'))
    

    def test_duplicate_fields_handling(self):
        """Test that duplicate fields in the fields list should raise ValueError"""
        class Parent(BaseModel):
            id: int
            name: str
            age: int
        
        with pytest.raises(ValueError):
            create_subset(Parent, ['id', 'name', 'id', 'name'], 'DuplicateSubset')
        
    
    def test_field_order_preservation(self):
        """Test that field order is preserved in subset."""
        class Parent(BaseModel):
            c: str
            a: int
            b: float
        
        Subset = create_subset(Parent, ['b', 'a', 'c'], 'OrderedSubset')
        
        instance = Subset(b=1.5, a=1, c='test')
        assert instance.a == 1
        assert instance.b == 1.5
        assert instance.c == 'test'
        
        # Check field order in model definition
        field_names = list(Subset.model_fields.keys())
        
        assert field_names == ['b', 'a', 'c']
    
    def test_nonexistent_field_error(self):
        """Test that referencing non-existent fields raises an error."""
        class Parent(BaseModel):
            id: int
            name: str
        
        with pytest.raises(AttributeError, match='field "nonexistent" not existed'):
            create_subset(Parent, ['id', 'nonexistent'], 'ErrorSubset')
    
    def test_non_basemodel_parent_error(self):
        """Test that using non-BaseModel parent raises an error."""
        class NotBaseModel:
            id: int
            name: str
        
        with pytest.raises(TypeError, match='parent must be a pydantic BaseModel'):
            create_subset(NotBaseModel, ['id'], 'ErrorSubset')  # type: ignore
    
    def test_empty_fields_list(self):
        """Test subset creation with empty fields list."""
        class Parent(BaseModel):
            id: int
            name: str
        
        Subset = create_subset(Parent, [], 'EmptySubset')
        
        # Should be able to create instance with no fields
        Subset()
        
        # Should have no fields
        field_names = list(Subset.model_fields.keys())
        assert len(field_names) == 0
        
        # Extra fields should be ignored (not raise TypeError)
        instance_with_extra = Subset(id=1)  # id is ignored
        assert not hasattr(instance_with_extra, 'id')
    
    def test_custom_subset_name(self):
        """Test that custom subset name is used correctly."""
        class Parent(BaseModel):
            id: int
            name: str
        
        Subset = create_subset(Parent, ['id'], 'CustomName')
        
        assert Subset.__name__ == 'CustomName'
    
    def test_default_subset_name(self):
        """Test default subset name when none provided."""
        class Parent(BaseModel):
            id: int
            name: str
        
        Subset = create_subset(Parent, ['id'])
        
        assert Subset.__name__ == 'SubsetModel'
    
    def test_model_configuration_inheritance(self):
        """Test that parent model configuration is inherited."""
        from pydantic import ConfigDict
        
        class Parent(BaseModel):
            model_config = ConfigDict(str_strip_whitespace=True, validate_assignment=True)
            name: str
            value: int
        
        Subset = create_subset(Parent, ['name'], 'ConfigSubset')
        
        # Test that config is inherited (behavior may vary based on implementation)
        instance = Subset(name='  test  ')
        assert instance.name == 'test'
        
        # The actual config inheritance testing would depend on the implementation details


class TestCreateSubsetIntegration:
    """Integration tests for create_subset with complex scenarios."""
    
    def test_nested_model_fields(self):
        """Test subset creation with nested model fields."""
        class Address(BaseModel):
            street: str
            city: str
        
        class Person(BaseModel):
            id: int
            name: str
            address: Address
            age: int
        
        Subset = create_subset(Person, ['id', 'address'], 'NestedSubset')
        
        address = Address(street='123 Main St', city='Anytown')
        instance = Subset(id=1, address=address)
        
        assert instance.id == 1
        assert instance.address.street == '123 Main St'
        assert instance.address.city == 'Anytown'
    
    def test_multiple_subsets_from_same_parent(self):
        """Test creating multiple subsets from the same parent."""
        class Parent(BaseModel):
            id: int
            name: str
            email: str
            age: int
            active: bool
        
        Subset1 = create_subset(Parent, ['id', 'name'], 'Subset1')
        Subset2 = create_subset(Parent, ['email', 'age'], 'Subset2')
        Subset3 = create_subset(Parent, ['id', 'active'], 'Subset3')
        
        # Test that all subsets work independently
        s1 = Subset1(id=1, name='test')
        s2 = Subset2(email='test@example.com', age=25)
        s3 = Subset3(id=2, active=True)
        
        assert s1.id == 1 and s1.name == 'test'
        assert s2.email == 'test@example.com' and s2.age == 25
        assert s3.id == 2 and s3.active is True
        
        # Test that subsets are truly independent
        assert Subset1.__name__ == 'Subset1'
        assert Subset2.__name__ == 'Subset2'
        assert Subset3.__name__ == 'Subset3'


class TestSubsetMeta:
    """Test cases for SubsetMeta metaclass and Subset base class."""
    
    def test_basic_subset_metaclass(self):
        """Test basic usage of Subset metaclass to create subset classes."""
        class Parent(BaseModel):
            id: int
            name: str
            age: int
            email: str
        
        class MySubset(DefineSubset):
            __pydantic_resolve_subset__ = (Parent, ['id', 'name'])
            new_field: str
        
        # Test that MySubset is a proper subset
        instance = MySubset(id=1, name='test', new_field='extra')
        assert instance.id == 1
        assert instance.name == 'test'
        assert instance.new_field == 'extra'

        # Test that excluded fields are not present
        field_names = list(MySubset.model_fields.keys())
        assert 'id' in field_names
        assert 'name' in field_names
        assert 'age' not in field_names
        assert 'email' not in field_names
        
        # Test that MySubset is a subclass of BaseModel
        assert issubclass(MySubset, BaseModel)
        assert getattr(MySubset, const.ENSURE_SUBSET_REFERENCE) is Parent
    
    def test_basic_subset_metaclass_with_fields_in_tuple(self):
        """Test basic usage of Subset metaclass to create subset classes."""
        class Parent(BaseModel):
            id: int
            name: str
            age: int
            email: str
        
        class MySubset(DefineSubset):
            __subset__ = (Parent, ('id', 'name'))
            new_field: str
        
        # Test that MySubset is a proper subset
        instance = MySubset(id=1, name='test', new_field='extra')
        assert instance.id == 1
        assert instance.name == 'test'
        assert instance.new_field == 'extra'
        
        # Test that MySubset is a subclass of BaseModel
        assert issubclass(MySubset, BaseModel)
        assert getattr(MySubset, const.ENSURE_SUBSET_REFERENCE) is Parent
    
    def test_wrongly_create_subset_metaclass(self):
        """Test basic usage of Subset metaclass to create subset classes."""
        class Parent(BaseModel):
            id: int
            name: str
            age: int
            email: str
        
        with pytest.raises(ValueError):
            class MySubset(DefineSubset):
                __pydantic_resolve_subset__ = (Parent, ['id', 'name'])
                id: str
        
        with pytest.raises(ValueError):
            class MySubset2(DefineSubset):
                id: str
        

class TestSubsetResolve:
    @pytest.mark.asyncio
    async def test_basic_subset_creation(self):
        class Parent(BaseModel):
            id: int
            name: str
            age: int
            email: str

        class Item(BaseModel):
            __pydantic_resolve_collect__ = {
                'count': 'collector'
            }
            name: str = ''
            def resolve_name(self, ancestor_context):
                return ancestor_context['my_subset_name']
            
            count: int
        
        class MySubset(DefineSubset):
            __pydantic_resolve_subset__ = (Parent, ('id', 'name'))
            __pydantic_resolve_expose__ = {
                'name': 'my_subset_name'
            }
            new_field: str = ''

            def resolve_new_field(self) -> str:
                return f"{self.id}-{self.name}"
            
            items: list[Item] = []
            def resolve_items(self):
                return [Item(count=1), Item(count=2)] 
            
            total: int = 0
            def post_total(self, collector=Collector(alias='collector')):
                return sum(collector.values())
            
            hello: str = 'world'
            def post_default_handler(self):
                self.hello = 'hello world'
        
        instance = MySubset(id=1, name='test')
        await Resolver().resolve(instance)
        assert instance.new_field == '1-test'
        assert instance.items[0].name == 'test'
        assert instance.total == 3
        assert instance.hello == 'hello world'
        

class TestCreateSubsetConfig:
    
    def test_case_1(self):
        class Parent(BaseModel):
            id: int
            name: str
            email: str
            age: int
            active: bool
        
        class Sub(DefineSubset):
            __subset__ = SubsetConfig(
                kls=Parent,
                fields=['id', 'name'],
            )
        
        sub = Sub(id=1, name='test')
        assert sub.id == 1
        assert sub.name == 'test'
        
        
    def test_case_2(self):
        class Parent(BaseModel):
            id: int
            name: str
            email: str
            age: int
            active: bool
        
        class Sub(DefineSubset):
            __subset__ = SubsetConfig(
                kls=Parent,
                omit_fields=['email'],
            )
        
        sub = Sub(id=1, name='test', age=30, active=True)
        assert sub.id == 1
        assert sub.name == 'test'
        
    def test_case_3(self):
        class Parent(BaseModel):
            id: int
            name: str
            email: str
            age: int
            active: bool
        
        class Sub(DefineSubset):
            __subset__ = SubsetConfig(
                kls=Parent,
                fields=['id', 'name', 'age'],
                expose_as=[('name', 'custom_name')],
                send_to=[
                    ('age', 'age_collector'),
                    ('age', ('a', 'b')) 
                ],
            )
        
        sub = Sub(id=1, name='test', age=30, active=True)
        assert sub.id == 1
        assert sub.name == 'test'

        assert Sub.model_fields['name'].metadata[0].alias == 'custom_name'

    def test_case_4(self):
        class Parent(BaseModel):
            id: int
            name: str
            email: str
            age: int
            active: bool
        
        class Sub(DefineSubset):
            __subset__ = SubsetConfig(
                kls=Parent,
                fields=['id', 'name', 'age'],
                excluded_fields=['age']
            )
        
        sub = Sub(id=1, name='test', age=30)
        assert sub.id == 1
        assert sub.name == 'test'
        assert sub.age == 30
        
        # Check if exclude=True is set
        assert Sub.model_fields['age'].exclude is True
        assert Sub.model_fields['name'].exclude is None
        
        # Check serialization
        assert sub.model_dump() == {'id': 1, 'name': 'test'} 


    def test_case_5(self):
        class Parent(BaseModel):
            id: int
            name: str
            email: str
            age: int
            active: bool
        
        class Sub(DefineSubset):
            __subset__ = SubsetConfig(
                kls=Parent,
                fields="all"
            )
        
        assert Sub.model_fields.keys() == Parent.model_fields.keys()
        

    def test_case_6(self):
        class Parent(BaseModel):
            id: int
            name: str
            email: str
            age: int
            active: bool
        
        class Sub(DefineSubset):
            __subset__ = SubsetConfig(
                kls=Parent,
                omit_fields=[]  # same as fields="all"
            )
        
        assert Sub.model_fields.keys() == Parent.model_fields.keys()
        


import abc
from dataclasses import dataclass
from typing import Any, List, Union, Tuple, Iterator, Type
import pydantic_resolve.constant as const

@dataclass
class SendToInfo:
    collector_name: str | Tuple[str]

def SendTo(name: str| Tuple[str]) -> SendToInfo:
    return SendToInfo(collector_name=name)


def pre_generate_collector_config(kls):
    """
    iterrate kls fields, check and collect field who's annotated metadata for SendTo exists
    if kls's const.COLLECTOR_CONFIGURATION exists and the fields is not empty, raise exception
    group those field name based on collector_name, if single, leave it as str, else make it tuple
    then generate the configuration such as 
    { (field_a, field_b): collector_name } or ( field_a: collector_name })
    and set it into kls's const.COLLECTOR_CONFIGURATION
    """
    fields = list(_get_pydantic_field_items_with_send_to(kls))
    if not fields:
        return

    if hasattr(kls, const.COLLECTOR_CONFIGURATION):
        raise AttributeError(
            f"{const.COLLECTOR_CONFIGURATION} already exists; cannot use SendTo annotations at the same time"
        )

    grouped: dict[object, list[str]] = {}
    for field_name, meta in fields:
        grouped.setdefault(meta.collector_name, []).append(field_name)

    collect_dict: dict[object, object] = {}
    for collector_name, field_names in grouped.items():
        key: object = field_names[0] if len(field_names) == 1 else tuple(field_names)
        collect_dict[key] = collector_name

    setattr(kls, const.COLLECTOR_CONFIGURATION, collect_dict)

def _get_pydantic_field_items_with_send_to(kls) -> Iterator[Tuple[str, SendToInfo, Type]]:
    items = kls.model_fields.items()

    for name, v in items:
        metadata = v.metadata
        for meta in metadata:
            if isinstance(meta, SendToInfo):
                yield name, meta

class ICollector(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __init__(self, alias: str):
        self.alias = alias

    @abc.abstractmethod
    def add(self, val):
        """how to add new element(s)"""

    @abc.abstractmethod
    def values(self) -> Any:
        """get result"""


class Collector(ICollector):
    def __init__(self, alias: str, flat: bool=False):
        super().__init__(alias)
        self.flat = flat
        self.val = []

    def add(self, val: Union[Any, List[Any]]) -> None:
        if self.flat:
            if isinstance(val, list):
                self.val.extend(val)
            else:
                raise TypeError('if flat, target should be list')
        else:
            self.val.append(val)

    def values(self) -> List[Any]:
        return self.val


from __future__ import annotations
import pytest
from typing import List
from collections import Counter, defaultdict
from aiodataloader import DataLoader
from pydantic import ConfigDict, BaseModel
from sqlalchemy import select
from sqlalchemy.ext.asyncio import async_sessionmaker
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from pydantic_resolve import Resolver, LoaderDepend


"""
CAUTION: using annotations at top
so all schema should be defined in global scope
"""

counter = Counter()
engine = create_async_engine(
    "sqlite+aiosqlite://",
    echo=False,
)
async_session = async_sessionmaker(engine, expire_on_commit=False)

class Base(DeclarativeBase):
    pass

class Task(Base):
    __tablename__ = "task"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str]

class Comment(Base):
    __tablename__ = "comment"
    id: Mapped[int] = mapped_column(primary_key=True)
    task_id: Mapped[int] = mapped_column(index=True)
    content: Mapped[str]

class Feedback(Base):
    __tablename__ = "feedback"
    id: Mapped[int] = mapped_column(primary_key=True)
    comment_id: Mapped[int] = mapped_column(index=True)
    content: Mapped[str]
    private: Mapped[bool]

async def insert_objects() -> None:
    async with async_session() as session:
        async with session.begin():
            session.add_all(
                [
                    Task(id=1, name="task-1"),

                    Comment(id=1, task_id=1, content="comment-1 for task 1"),

                    Feedback(id=1, comment_id=1, content="feedback-1 for comment-1", private=True),
                    Feedback(id=2, comment_id=1, content="feedback-2 for comment-1", private=False),
                    Feedback(id=3, comment_id=1, content="feedback-3 for comment-1", private=True),
                ]
            )

async def insert_and_update_objects() -> None:
    async with async_session() as session:
        async with session.begin():
            task_1 = (await session.execute(select(Task).filter_by(id=1))).scalar_one()
            task_1.name = 'task-1 xyz'

            comment_1 = (await session.execute(select(Comment).filter_by(id=1))).scalar_one()
            comment_1.content = 'comment-1 for task 1 (changes)'

            feedback_1 = (await session.execute(select(Feedback).filter_by(id=1))).scalar_one()
            feedback_1.content = 'feedback-1 for comment-1 (changes)'

            session.add(task_1)
            session.add(comment_1)
            session.add(feedback_1)
            session.add_all(
                [
                    Comment(id=2, task_id=1, content="comment-2 for task 1"),

                    Feedback(id=4, comment_id=2, content="test", private=False),
                ]
            )


# =========================== Pydantic Schema layer =========================
class FeedbackLoader(DataLoader):
    private: bool
    async def batch_load_fn(self, comment_ids):
        counter['load-feedback'] += 1
        async with async_session() as session:
            res = await session.execute(select(Feedback).where(Feedback.private == self.private).where(Feedback.comment_id.in_(comment_ids)))
            rows = res.scalars().all()
            dct = defaultdict(list)
            for row in rows:
                dct[row.comment_id].append(FeedbackSchema.model_validate(row))
            return [dct.get(k, []) for k in comment_ids]

class CommentLoader(DataLoader):
    async def batch_load_fn(self, task_ids):
        counter['load-comment'] += 1
        async with async_session() as session:
            res = await session.execute(select(Comment).where(Comment.task_id.in_(task_ids)))
            rows = res.scalars().all()

            dct = defaultdict(list)
            for row in rows:
                dct[row.task_id].append(CommentSchema.model_validate(row))
            return [dct.get(k, []) for k in task_ids]

class FeedbackSchema(BaseModel):
    id: int
    comment_id: int
    content: str
    model_config = ConfigDict(from_attributes=True)

class CommentSchema(BaseModel):
    id: int
    task_id: int
    content: str
    feedbacks: List[FeedbackSchema]  = []

    def resolve_feedbacks(self, loader=LoaderDepend(FeedbackLoader)):
        return loader.load(self.id)
    model_config = ConfigDict(from_attributes=True)

class TaskSchema(BaseModel):
    id: int
    name: str
    comments: List[CommentSchema]  = []
    
    def resolve_comments(self, loader=LoaderDepend(CommentLoader)):
        return loader.load(self.id)
    model_config = ConfigDict(from_attributes=True)

async def init():
    async with engine.begin() as conn:
        await conn.run_sync(Base.metadata.create_all)
    await insert_objects()

async def task_query(task_id = 1):
    async with async_session() as session:
        tasks = (await session.execute(select(Task).where(Task.id == task_id))).scalars().all()
        task_objs = [TaskSchema.model_validate(t) for t in tasks]
        resolved_results = await Resolver(loader_params={FeedbackLoader: {'private': True}}).resolve(task_objs)
        to_dict_arr = [r.model_dump() for r in resolved_results]
        return to_dict_arr

@pytest.mark.asyncio
async def test_sqlite_and_dataloader():
    await init()
    result_1 = await task_query()
    await insert_and_update_objects()
    result_2 = await task_query()

    expected_1 = [
        {
            'id': 1,
            'name': 'task-1',
            'comments': [
                {
                    'content': 'comment-1 for task 1',
                    'feedbacks': [
                        {'comment_id': 1, 'content': 'feedback-1 for comment-1', 'id': 1},
                        # {'comment_id': 1, 'content': 'feedback-2 for comment-1', 'id': 2},
                        {'comment_id': 1, 'content': 'feedback-3 for comment-1', 'id': 3}
                        ],
                    'id': 1,
                    'task_id': 1},
                ],
            }
        ]

    expected_2 = [
        {
            'id': 1,
            'name': 'task-1 xyz',
            'comments': [
                {
                    'content': 'comment-1 for task 1 (changes)',
                    'feedbacks': [
                        {'comment_id': 1, 'content': 'feedback-1 for comment-1 (changes)', 'id': 1},
                        # {'comment_id': 1, 'content': 'feedback-2 for comment-1', 'id': 2},
                        {'comment_id': 1, 'content': 'feedback-3 for comment-1', 'id': 3}
                        ],
                    'id': 1,
                    'task_id': 1},
                {
                    'content': 'comment-2 for task 1',
                    'feedbacks': [
                        # {'comment_id': 2, 'content': 'test', 'id': 4},   # private is false
                    ],
                    'id': 2,
                    'task_id': 1}
                ],
            }
        ]

    assert result_1 == expected_1
    assert result_2 == expected_2
    
    assert counter['load-comment'] == 2  # Resolver + LoaderDepend can fix cache issue
    assert counter['load-feedback'] == 2



from __future__ import annotations
from typing import List
import pytest
from pydantic import BaseModel
from pydantic_resolve import Resolver, LoaderDepend

BOOKS = {
    1: [{'name': 'book1'}, {'name': 'book2'}],
    2: [{'name': 'book3'}, {'name': 'book4'}],
    3: [{'name': 'book1'}, {'name': 'book2'}],
}
async def batch_load_fn(keys):
    books = [BOOKS.get(k, []) for k in keys]
    return books

# in reverse order

class ClassRoom(BaseModel):
    students: List[Student] = []
    def resolve_students(self):
        students = [dict(id=1, name="jack"), dict(id=2, name="mike"), dict(id=3, name="wiki")]
        return students

class Student(BaseModel):
    id: int
    name: str

    books: List[Book] = [] 
    def resolve_books(self, loader=LoaderDepend(batch_load_fn)):
        return loader.load(self.id)

class Book(BaseModel):
    name: str


@pytest.mark.asyncio
async def test_1():
    classroom = ClassRoom()
    classroom = await Resolver().resolve(classroom)  # annotation is not required any more.
    assert isinstance(classroom.students[0].books[0], Book)




from .utils.collector import Collector, ICollector, SendTo
from .utils.class_util import ensure_subset
from .utils.dataloader import build_list, build_object, copy_dataloader_kls
from .utils.conversion import mapper
from .exceptions import (
    ResolverTargetAttrNotFound,
    LoaderFieldNotProvidedError,
    MissingAnnotationError,
    GlobalLoaderFieldOverlappedError,
    MissingCollector)
from .resolver import Resolver
from .utils.depend import LoaderDepend, Loader
from .utils.subset import DefineSubset, SubsetConfig
from .utils.openapi import (
    model_config)
from .utils.er_diagram import Relationship, MultipleRelationship, Link, Entity, ErDiagram, LoadBy, base_entity
from .utils.resolver_configurator import config_resolver, config_global_resolver
from .utils.expose import ExposeAs


__all__ = [
    'Resolver',
    'LoaderDepend',
    'Loader',  # short 
    'Collector',
    'ICollector',
    'ExposeAs',
    'SendTo',

    'ResolverTargetAttrNotFound',
    'LoaderFieldNotProvidedError',
    'MissingAnnotationError',
    'GlobalLoaderFieldOverlappedError',
    'MissingCollector',

    'build_list',
    'build_object',
    'mapper',
    'ensure_subset',
    'model_config',
    'copy_dataloader_kls',
    'DefineSubset',
    'SubsetConfig',

    'Entity',
    'Relationship',
    'MultipleRelationship',
    'Link',
    'ErDiagram',
    'LoadBy',
    'base_entity',

    'config_resolver',
    'config_global_resolver',
]


import pytest
from typing import Optional, Annotated
from pydantic import BaseModel
from pydantic_resolve import config_resolver
from pydantic_resolve import Entity, Relationship, LoadBy, ErDiagram
from aiodataloader import DataLoader


class Biz(BaseModel):
    id: Optional[int]
    name: str
    user_id: Optional[int]

class Bar(BaseModel):
    id: int
    name: str
    biz_id: int

class User(BaseModel):
    id: int
    name: str

class UserLoader(DataLoader):
    async def batch_load_fn(self, keys):
        users = [
            dict(id=1, name="a"),
            dict(id=2, name="b"),
            dict(id=3, name="c"),
        ]
        user_map = { u['id']: u for u in users}
        return [user_map.get(k, None) for k in keys]

class BarLoader(DataLoader):
    async def batch_load_fn(self, keys):
        bars = [
            dict(id=1, name="bar1", biz_id=1),
            dict(id=2, name="bar2", biz_id=1),
            dict(id=3, name="bar3", biz_id=2),
        ]
        bar_map = {}
        for b in bars:
            bar_map.setdefault(b['biz_id'], []).append(b)
        return [bar_map.get(k, []) for k in keys]

diagram = ErDiagram(
    configs=[
        Entity(kls=Biz, relationships=[
            Relationship(field='user_id', target_kls=User, loader=UserLoader),
            Relationship(field='id', target_kls=list[Bar], field_none_default_factory=list, loader=BarLoader),
        ])
    ]
)

class BizCase1(Biz):
    user: Annotated[Optional[User], LoadBy('user_id')] = None
    bars: Annotated[list[Bar], LoadBy('id')] = []
    

@pytest.mark.asyncio
async def test_resolver_factory_with_er_configs_inherit():
    MyResolver = config_resolver('MyResolver', er_diagram=diagram)
    d = [BizCase1(id=None, name="qq", user_id=None)]
    d = await MyResolver().resolve(d)

    assert d[0].user is None
    assert d[0].bars == []


from pydantic import BaseModel
from pydantic_resolve import Resolver
import pytest

class A(BaseModel):
    id: int
    name: str

    desc: str = ''
    def post_desc(self):
        return f'a: {self.name}-{self.id}'


class B(BaseModel):
    id: int
    name: str

    desc: str = ''
    def post_desc(self):
        return f'b: {self.name}-{self.id}'


Item = A | B

@pytest.mark.asyncio
async def test_annotation_param():
    items: list[Item] = [
        A(id=1, name='item1'),
        B(id=2, name='item2'),
        A(id=3, name='item3'),
    ]

    resolved_items = await Resolver(annotation=Item).resolve(items)

    assert resolved_items[0].desc == 'a: item1-1'
    assert resolved_items[1].desc == 'b: item2-2'
    assert resolved_items[2].desc == 'a: item3-3'

@pytest.mark.asyncio
async def test_without_annotation_param():
    items: list[Item] = [
        A(id=1, name='item1'),
        B(id=2, name='item2'),
        A(id=3, name='item3'),
    ]
    with pytest.raises(Exception):
        await Resolver().resolve(items)


from typing import Union, List
from pydantic import BaseModel
from pydantic_resolve import Resolver, Loader
import pytest


async def batch_load_fn(keys):
    return [1 for k in keys]

class A(BaseModel):
    id: str
    def post_id(self):
        return self.id + "-"

    n: int = 0
    def resolve_n(self, loader=Loader(batch_load_fn)):
        return loader.load(self.id)

class B(BaseModel):
    id: str
    name: str
    def post_id(self):
        return self.id + "+"

    n: int = 0
    def resolve_n(self, loader=Loader(batch_load_fn)):
        return loader.load(self.id)

class C(BaseModel):
    id: str
    name: str


Item = Union[A, B, C]


class Container(BaseModel):
    items: List[Item] = []

    def resolve_items(self):
        return [A(id='1'), B(id='2', name='Item 2')]


@pytest.mark.asyncio
async def test_type_definition():
    c = Container()
    resolver = Resolver()
    result = await resolver.resolve(c)
    expected = {
        'items': [
            {'id': '1-', 'n': 1},
            {'id': '2+', 'name': 'Item 2', 'n': 1}
        ],
    }
    assert result.model_dump() == expected



from __future__ import annotations
import pytest
from typing import Optional, List
from pydantic import BaseModel
from aiodataloader import DataLoader
from pydantic_resolve.analysis import Analytic
from pydantic_resolve import LoaderDepend, LoaderFieldNotProvidedError
from pydantic_resolve.loader_manager import validate_and_create_loader_instance

async def loader_fn(keys):
    return keys

class MyLoader(DataLoader):
    param: str
    async def batch_load_fn(self, keys):
        return keys

class Student(BaseModel):
    __pydantic_resolve_expose__ = {'name': 'student_name'}
    name: str = ''
    resolve_hello: str = ''

    def resolve_name(self, context, ancestor_context, loader=LoaderDepend(loader_fn)):
        return '.'

    def post_name(self, loader=LoaderDepend(loader_fn)):
        return '.'

    zones: List[Optional[Zone]] = [None]
    zones2: List[Zone] = []
    zone: Optional[Optional[Zone]] = None

    zeta: Optional[Zeta] = None

    def resolve_zeta(self):
        return dict(name='z')

class Zone(BaseModel):
    name: str
    qs: List[Queue]
    def resolve_qs(self, qs_loader=LoaderDepend(MyLoader)):
        return qs_loader.load(self.name)

class Queue(BaseModel):
    name: str

class Zeta(BaseModel):
    name: str


loader_params = {
    MyLoader: {
        'param': 'aaa'
    }
}
global_loader_param = {
    'param': 'aaa'
}

@pytest.mark.asyncio
async def test_instance_1():
    metadata = Analytic().scan(Student)
    loader_instance = validate_and_create_loader_instance(loader_params, {}, {}, metadata)

    assert isinstance(loader_instance['test_field_validate_and_create_loader_instance.loader_fn'] , DataLoader)
    assert isinstance(loader_instance['test_field_validate_and_create_loader_instance.MyLoader'] , MyLoader)

@pytest.mark.asyncio
async def test_instance_2():
    """ test cache works """
    metadata = Analytic().scan(Student)
    loader_instance = validate_and_create_loader_instance(loader_params, {}, {}, metadata)

    assert len(loader_instance) == 2

@pytest.mark.asyncio
async def test_instance_3():
    """test global param"""
    metadata = Analytic().scan(Student)
    loader_instance = validate_and_create_loader_instance({}, global_loader_param, {}, metadata)

    assert isinstance(loader_instance['test_field_validate_and_create_loader_instance.loader_fn'] , DataLoader)
    assert len(loader_instance) == 2

@pytest.mark.asyncio
async def test_instance_4():
    """raise missing param error"""
    metadata = Analytic().scan(Student)

    with pytest.raises(LoaderFieldNotProvidedError):
        validate_and_create_loader_instance({}, {}, {}, metadata)


from pydantic import BaseModel, Field
from typing import Annotated
from pydantic_resolve.utils.er_diagram import LoadBy

class User(BaseModel):
    age: Annotated[int, Field(ge=0, le=120, description="用户年龄"), LoadBy('user_id')] = 18

for k, v in User.model_fields.items():
    print(f"Field: {k}, Info: {v}" )
    print(v.metadata[2])





from pydantic import BaseModel
from pydantic_resolve.analysis import _scan_post_method, _scan_post_default_handler
from pydantic_resolve import Collector

def test_scan_post_method_1():
    class A(BaseModel):
        a: str
        def post_a(self):
            return 2 * self.a
        
    result = _scan_post_method(A.post_a, 'post_a', None)

    assert result == {
        'trim_field': 'a',
        'context': False,
        'ancestor_context': False,
        'parent': False,
        'dataloaders': [],
        'collectors': []
    }


def test_scan_post_method_2():
    class A(BaseModel):
        a: str
        def post_a(self, context, ancestor_context, parent):
            return 2 * self.a
        
    result = _scan_post_method(A.post_a, 'post_a', None)

    assert result == {
        'trim_field': 'a',
        'context': True,
        'ancestor_context': True,
        'parent': True,
        'dataloaders': [],
        'collectors': []
    }


def test_scan_post_method_3():
    class A(BaseModel):
        a: str
        def post_a(self,
                   context,
                   ancestor_context,
                   collector=Collector(alias='c_name'),
                   collector_2=Collector(alias='c_name'),
                   ):
            return 2 * self.a
        
    result = _scan_post_method(A.post_a, 'post_a', None)

    assert len(result['collectors']) == 2
    assert result['collectors'][0]['field'] == 'post_a' 
    assert result['collectors'][0]['param'] == 'collector' 
    assert result['collectors'][0]['alias'] == 'c_name'
    assert isinstance(result['collectors'][0]['instance'], Collector)

    assert result['collectors'][1]['field'] == 'post_a' 
    assert result['collectors'][1]['param'] == 'collector_2' 
    assert result['collectors'][1]['alias'] == 'c_name'
    assert isinstance(result['collectors'][1]['instance'], Collector)

def test_scan_post_method_4():
    class A(BaseModel):
        a: str
        def post_a(self, context, ancestor_context, collector=Collector(alias='c_name')):
            return 2 * self.a
        
        def post_default_handler(self, context, ancestor_context, parent):
            return 1
        
    result = _scan_post_default_handler(A.post_default_handler)

    assert result['context']
    assert result['parent']
    assert result['ancestor_context']




from pydantic_resolve.utils.openapi import model_config
from pydantic import BaseModel, Field
from pydantic_resolve import Resolver
from pydantic.json_schema import GenerateJsonSchema as GenerateJsonSchema
import pytest
from typing import Annotated

@pytest.mark.asyncio
async def test_schema_config():

    @model_config()
    class Y(BaseModel):
        id: int = Field(default=0, exclude=True)
        def resolve_id(self):
            return 1

        name: str

        password: str = Field(default='', exclude=True)
        def resolve_password(self):
            return 'confidential'

    schema = Y.model_json_schema()
    assert list((schema['properties']).keys()) == ['name']

    y = Y(name='kikodo')

    y = await Resolver().resolve(y) 
    assert y.model_dump() == {'name': 'kikodo'}



@pytest.mark.asyncio
async def test_schema_config_required_false():
    @model_config(default_required=False)
    class Y(BaseModel):
        id: int = 0
        def resolve_id(self):
            return 1

        name: str

        password: str = ''
        def resolve_password(self):
            return 'confidential'

    schema = Y.model_json_schema()
    assert set(schema['required']) == {'name'}


@pytest.mark.asyncio
async def test_schema_config_required_true():
    @model_config(default_required=True)
    class Y(BaseModel):
        id: Annotated[int, 'hello'] = 0

        name: str

        password: str = ''
        def resolve_password(self):
            return 'confidential'

    schema = Y.model_json_schema()
    assert set(schema['required']) == {'name', 'id', 'password'}


@pytest.mark.asyncio
async def test_raw_schema_in_serialization():
    """
    in pydantic v2 and fastapi, it will generate json schema in mode: serialization
    it will ignore fields with exclude=True, but model_config still helps to set required fields
    """
    class Y(BaseModel):
        name: str
        id: int = 0
        password: str = Field(default='', exclude=True)

    schema = Y.model_json_schema(mode='serialization')
    assert list((schema['properties']).keys()) == ['name', 'id']
    assert set(schema['required']) == {'name'}
